"""
This type stub file was generated by pyright.
"""

import numpy as np
from typing import Hashable, Literal
from pandas._typing import npt

def unique_label_indices(labels: np.ndarray) -> np.ndarray:
    ...

class Factorizer:
    count: int
    def __init__(self, size_hint: int) -> None:
        ...
    
    def get_count(self) -> int:
        ...
    


class ObjectFactorizer(Factorizer):
    table: PyObjectHashTable
    uniques: ObjectVector
    def factorize(self, values: npt.NDArray[np.object_], sort: bool = ..., na_sentinel=..., na_value=...) -> npt.NDArray[np.intp]:
        ...
    


class Int64Factorizer(Factorizer):
    table: Int64HashTable
    uniques: Int64Vector
    def factorize(self, values: np.ndarray, sort: bool = ..., na_sentinel=..., na_value=...) -> npt.NDArray[np.intp]:
        ...
    


class Int64Vector:
    def __init__(self, *args) -> None:
        ...
    
    def __len__(self) -> int:
        ...
    
    def to_array(self) -> npt.NDArray[np.int64]:
        ...
    


class Int32Vector:
    def __init__(self, *args) -> None:
        ...
    
    def __len__(self) -> int:
        ...
    
    def to_array(self) -> npt.NDArray[np.int32]:
        ...
    


class Int16Vector:
    def __init__(self, *args) -> None:
        ...
    
    def __len__(self) -> int:
        ...
    
    def to_array(self) -> npt.NDArray[np.int16]:
        ...
    


class Int8Vector:
    def __init__(self, *args) -> None:
        ...
    
    def __len__(self) -> int:
        ...
    
    def to_array(self) -> npt.NDArray[np.int8]:
        ...
    


class UInt64Vector:
    def __init__(self, *args) -> None:
        ...
    
    def __len__(self) -> int:
        ...
    
    def to_array(self) -> npt.NDArray[np.uint64]:
        ...
    


class UInt32Vector:
    def __init__(self, *args) -> None:
        ...
    
    def __len__(self) -> int:
        ...
    
    def to_array(self) -> npt.NDArray[np.uint32]:
        ...
    


class UInt16Vector:
    def __init__(self, *args) -> None:
        ...
    
    def __len__(self) -> int:
        ...
    
    def to_array(self) -> npt.NDArray[np.uint16]:
        ...
    


class UInt8Vector:
    def __init__(self, *args) -> None:
        ...
    
    def __len__(self) -> int:
        ...
    
    def to_array(self) -> npt.NDArray[np.uint8]:
        ...
    


class Float64Vector:
    def __init__(self, *args) -> None:
        ...
    
    def __len__(self) -> int:
        ...
    
    def to_array(self) -> npt.NDArray[np.float64]:
        ...
    


class Float32Vector:
    def __init__(self, *args) -> None:
        ...
    
    def __len__(self) -> int:
        ...
    
    def to_array(self) -> npt.NDArray[np.float32]:
        ...
    


class Complex128Vector:
    def __init__(self, *args) -> None:
        ...
    
    def __len__(self) -> int:
        ...
    
    def to_array(self) -> npt.NDArray[np.complex128]:
        ...
    


class Complex64Vector:
    def __init__(self, *args) -> None:
        ...
    
    def __len__(self) -> int:
        ...
    
    def to_array(self) -> npt.NDArray[np.complex64]:
        ...
    


class StringVector:
    def __init__(self, *args) -> None:
        ...
    
    def __len__(self) -> int:
        ...
    
    def to_array(self) -> npt.NDArray[np.object_]:
        ...
    


class ObjectVector:
    def __init__(self, *args) -> None:
        ...
    
    def __len__(self) -> int:
        ...
    
    def to_array(self) -> npt.NDArray[np.object_]:
        ...
    


class HashTable:
    def __init__(self, size_hint: int = ...) -> None:
        ...
    
    def __len__(self) -> int:
        ...
    
    def __contains__(self, key: Hashable) -> bool:
        ...
    
    def sizeof(self, deep: bool = ...) -> int:
        ...
    
    def get_state(self) -> dict[str, int]:
        ...
    
    def get_item(self, item):
        ...
    
    def set_item(self, item) -> None:
        ...
    
    def map_locations(self, values: np.ndarray) -> None:
        ...
    
    def lookup(self, values: np.ndarray) -> npt.NDArray[np.intp]:
        ...
    
    def get_labels(self, values: np.ndarray, uniques, count_prior: int = ..., na_sentinel: int = ..., na_value: object = ...) -> npt.NDArray[np.intp]:
        ...
    
    def unique(self, values: np.ndarray, return_inverse: bool = ...) -> tuple[np.ndarray, npt.NDArray[np.intp],] | np.ndarray:
        ...
    
    def factorize(self, values: np.ndarray, na_sentinel: int = ..., na_value: object = ..., mask=...) -> tuple[np.ndarray, npt.NDArray[np.intp]]:
        ...
    


class Complex128HashTable(HashTable):
    ...


class Complex64HashTable(HashTable):
    ...


class Float64HashTable(HashTable):
    ...


class Float32HashTable(HashTable):
    ...


class Int64HashTable(HashTable):
    def get_labels_groupby(self, values: npt.NDArray[np.int64]) -> tuple[npt.NDArray[np.intp], npt.NDArray[np.int64]]:
        ...
    
    def map_keys_to_values(self, keys: npt.NDArray[np.int64], values: npt.NDArray[np.int64]) -> None:
        ...
    


class Int32HashTable(HashTable):
    ...


class Int16HashTable(HashTable):
    ...


class Int8HashTable(HashTable):
    ...


class UInt64HashTable(HashTable):
    ...


class UInt32HashTable(HashTable):
    ...


class UInt16HashTable(HashTable):
    ...


class UInt8HashTable(HashTable):
    ...


class StringHashTable(HashTable):
    ...


class PyObjectHashTable(HashTable):
    ...


class IntpHashTable(HashTable):
    ...


def duplicated(values: np.ndarray, keep: Literal["last", "first", False] = ...) -> npt.NDArray[np.bool_]:
    ...

def mode(values: np.ndarray, dropna: bool, mask: npt.NDArray[np.bool_] | None = ...) -> np.ndarray:
    ...

def value_count(values: np.ndarray, dropna: bool, mask: npt.NDArray[np.bool_] | None = ...) -> tuple[np.ndarray, npt.NDArray[np.int64]]:
    ...

def ismember(arr: np.ndarray, values: np.ndarray) -> npt.NDArray[np.bool_]:
    ...

def object_hash(obj) -> int:
    ...

def objects_are_equal(a, b) -> bool:
    ...




"""
This type stub file was generated by pyright.
"""

from pathlib import Path
from typing import Sequence, TYPE_CHECKING
from pandas import DataFrame

if TYPE_CHECKING:
    ...
def read_spss(path: str | Path, usecols: Sequence[str] | None = ..., convert_categoricals: bool = ...) -> DataFrame:
    """
    Load an SPSS file from the file path, returning a DataFrame.

    .. versionadded:: 0.25.0

    Parameters
    ----------
    path : str or Path
        File path.
    usecols : list-like, optional
        Return a subset of the columns. If None, return all columns.
    convert_categoricals : bool, default is True
        Convert categorical columns into pd.Categorical.

    Returns
    -------
    DataFrame
    """
    ...




"""
This type stub file was generated by pyright.
"""

from typing import Callable, TYPE_CHECKING, TypeVar
from pandas.util._decorators import doc
from pandas.core.indexes.base import Index
from pandas.core.arrays import IntervalArray
from pandas.core.arrays._mixins import NDArrayBackedExtensionArray

"""
Shared methods for Index subclasses backed by ExtensionArray.
"""
if TYPE_CHECKING:
    ...
_T = TypeVar("_T", bound="NDArrayBackedExtensionIndex")
_ExtensionIndexT = TypeVar("_ExtensionIndexT", bound="ExtensionIndex")
def inherit_names(names: list[str], delegate: type, cache: bool = ..., wrap: bool = ...) -> Callable[[type[_ExtensionIndexT]], type[_ExtensionIndexT]]:
    """
    Class decorator to pin attributes from an ExtensionArray to a Index subclass.

    Parameters
    ----------
    names : List[str]
    delegate : class
    cache : bool, default False
    wrap : bool, default False
        Whether to wrap the inherited result in an Index.
    """
    ...

class ExtensionIndex(Index):
    """
    Index subclass for indexes backed by ExtensionArray.
    """
    _data: IntervalArray | NDArrayBackedExtensionArray
    @doc(Index.map)
    def map(self, mapper, na_action=...): # -> Index | MultiIndex:
        ...
    


class NDArrayBackedExtensionIndex(ExtensionIndex):
    """
    Index subclass for indexes backed by NDArrayBackedExtensionArray.
    """
    _data: NDArrayBackedExtensionArray
    ...





"""
This type stub file was generated by pyright.
"""

import io
from typing import Sequence
from pandas._typing import CompressionOptions, ConvertersArg, DtypeArg, FilePath, ParseDatesArg, ReadBuffer, StorageOptions, TYPE_CHECKING, XMLParsers
from pandas.util._decorators import deprecate_nonkeyword_arguments, doc
from pandas.core.shared_docs import _shared_docs
from pandas import DataFrame

"""
:mod:`pandas.io.xml` is a module for reading XML.
"""
if TYPE_CHECKING:
    ...
@doc(storage_options=_shared_docs["storage_options"], decompression_options=_shared_docs["decompression_options"] % "path_or_buffer")
class _XMLFrameParser:
    """
    Internal subclass to parse XML into DataFrames.

    Parameters
    ----------
    path_or_buffer : a valid JSON str, path object or file-like object
        Any valid string path is acceptable. The string could be a URL. Valid
        URL schemes include http, ftp, s3, and file.

    xpath : str or regex
        The XPath expression to parse required set of nodes for
        migration to `Data Frame`. `etree` supports limited XPath.

    namespaces : dict
        The namespaces defined in XML document (`xmlns:namespace='URI')
        as dicts with key being namespace and value the URI.

    elems_only : bool
        Parse only the child elements at the specified `xpath`.

    attrs_only : bool
        Parse only the attributes at the specified `xpath`.

    names : list
        Column names for Data Frame of parsed XML data.

    dtype : dict
        Data type for data or columns. E.g. {{'a': np.float64,
        'b': np.int32, 'c': 'Int64'}}

        .. versionadded:: 1.5.0

    converters : dict, optional
        Dict of functions for converting values in certain columns. Keys can
        either be integers or column labels.

        .. versionadded:: 1.5.0

    parse_dates : bool or list of int or names or list of lists or dict
        Converts either index or select columns to datetimes

        .. versionadded:: 1.5.0

    encoding : str
        Encoding of xml object or document.

    stylesheet : str or file-like
        URL, file, file-like object, or a raw string containing XSLT,
        `etree` does not support XSLT but retained for consistency.

    iterparse : dict, optional
        Dict with row element as key and list of descendant elements
        and/or attributes as value to be retrieved in iterparsing of
        XML document.

        .. versionadded:: 1.5.0

    {decompression_options}

        .. versionchanged:: 1.4.0 Zstandard support.

    {storage_options}

    See also
    --------
    pandas.io.xml._EtreeFrameParser
    pandas.io.xml._LxmlFrameParser

    Notes
    -----
    To subclass this class effectively you must override the following methods:`
        * :func:`parse_data`
        * :func:`_parse_nodes`
        * :func:`_iterparse_nodes`
        * :func:`_parse_doc`
        * :func:`_validate_names`
        * :func:`_validate_path`


    See each method's respective documentation for details on their
    functionality.
    """
    def __init__(self, path_or_buffer: FilePath | ReadBuffer[bytes] | ReadBuffer[str], xpath: str, namespaces: dict[str, str] | None, elems_only: bool, attrs_only: bool, names: Sequence[str] | None, dtype: DtypeArg | None, converters: ConvertersArg | None, parse_dates: ParseDatesArg | None, encoding: str | None, stylesheet: FilePath | ReadBuffer[bytes] | ReadBuffer[str] | None, iterparse: dict[str, list[str]] | None, compression: CompressionOptions, storage_options: StorageOptions) -> None:
        ...
    
    def parse_data(self) -> list[dict[str, str | None]]:
        """
        Parse xml data.

        This method will call the other internal methods to
        validate xpath, names, parse and return specific nodes.
        """
        ...
    


class _EtreeFrameParser(_XMLFrameParser):
    """
    Internal class to parse XML into DataFrames with the Python
    standard library XML module: `xml.etree.ElementTree`.
    """
    def parse_data(self) -> list[dict[str, str | None]]:
        ...
    


class _LxmlFrameParser(_XMLFrameParser):
    """
    Internal class to parse XML into DataFrames with third-party
    full-featured XML library, `lxml`, that supports
    XPath 1.0 and XSLT 1.0.
    """
    def parse_data(self) -> list[dict[str, str | None]]:
        """
        Parse xml data.

        This method will call the other internal methods to
        validate xpath, names, optionally parse and run XSLT,
        and parse original or transformed XML and return specific nodes.
        """
        ...
    


def get_data_from_filepath(filepath_or_buffer: FilePath | bytes | ReadBuffer[bytes] | ReadBuffer[str], encoding: str | None, compression: CompressionOptions, storage_options: StorageOptions) -> str | bytes | ReadBuffer[bytes] | ReadBuffer[str]:
    """
    Extract raw XML data.

    The method accepts three input types:
        1. filepath (string-like)
        2. file-like object (e.g. open file object, StringIO)
        3. XML string or bytes

    This method turns (1) into (2) to simplify the rest of the processing.
    It returns input types (2) and (3) unchanged.
    """
    ...

def preprocess_data(data) -> io.StringIO | io.BytesIO:
    """
    Convert extracted raw data.

    This method will return underlying data of extracted XML content.
    The data either has a `read` attribute (e.g. a file object or a
    StringIO/BytesIO) or is a string or bytes that is an XML document.
    """
    ...

@deprecate_nonkeyword_arguments(version=None, allowed_args=["path_or_buffer"])
@doc(storage_options=_shared_docs["storage_options"], decompression_options=_shared_docs["decompression_options"] % "path_or_buffer")
def read_xml(path_or_buffer: FilePath | ReadBuffer[bytes] | ReadBuffer[str], xpath: str = ..., namespaces: dict[str, str] | None = ..., elems_only: bool = ..., attrs_only: bool = ..., names: Sequence[str] | None = ..., dtype: DtypeArg | None = ..., converters: ConvertersArg | None = ..., parse_dates: ParseDatesArg | None = ..., encoding: str | None = ..., parser: XMLParsers = ..., stylesheet: FilePath | ReadBuffer[bytes] | ReadBuffer[str] | None = ..., iterparse: dict[str, list[str]] | None = ..., compression: CompressionOptions = ..., storage_options: StorageOptions = ...) -> DataFrame:
    r"""
    Read XML document into a ``DataFrame`` object.

    .. versionadded:: 1.3.0

    Parameters
    ----------
    path_or_buffer : str, path object, or file-like object
        String, path object (implementing ``os.PathLike[str]``), or file-like
        object implementing a ``read()`` function. The string can be any valid XML
        string or a path. The string can further be a URL. Valid URL schemes
        include http, ftp, s3, and file.

    xpath : str, optional, default './\*'
        The XPath to parse required set of nodes for migration to DataFrame.
        XPath should return a collection of elements and not a single
        element. Note: The ``etree`` parser supports limited XPath
        expressions. For more complex XPath, use ``lxml`` which requires
        installation.

    namespaces : dict, optional
        The namespaces defined in XML document as dicts with key being
        namespace prefix and value the URI. There is no need to include all
        namespaces in XML, only the ones used in ``xpath`` expression.
        Note: if XML document uses default namespace denoted as
        `xmlns='<URI>'` without a prefix, you must assign any temporary
        namespace prefix such as 'doc' to the URI in order to parse
        underlying nodes and/or attributes. For example, ::

            namespaces = {{"doc": "https://example.com"}}

    elems_only : bool, optional, default False
        Parse only the child elements at the specified ``xpath``. By default,
        all child elements and non-empty text nodes are returned.

    attrs_only :  bool, optional, default False
        Parse only the attributes at the specified ``xpath``.
        By default, all attributes are returned.

    names :  list-like, optional
        Column names for DataFrame of parsed XML data. Use this parameter to
        rename original element names and distinguish same named elements and
        attributes.

    dtype : Type name or dict of column -> type, optional
        Data type for data or columns. E.g. {{'a': np.float64, 'b': np.int32,
        'c': 'Int64'}}
        Use `str` or `object` together with suitable `na_values` settings
        to preserve and not interpret dtype.
        If converters are specified, they will be applied INSTEAD
        of dtype conversion.

        .. versionadded:: 1.5.0

    converters : dict, optional
        Dict of functions for converting values in certain columns. Keys can either
        be integers or column labels.

        .. versionadded:: 1.5.0

    parse_dates : bool or list of int or names or list of lists or dict, default False
        Identifiers to parse index or columns to datetime. The behavior is as follows:

        * boolean. If True -> try parsing the index.
        * list of int or names. e.g. If [1, 2, 3] -> try parsing columns 1, 2, 3
          each as a separate date column.
        * list of lists. e.g.  If [[1, 3]] -> combine columns 1 and 3 and parse as
          a single date column.
        * dict, e.g. {{'foo' : [1, 3]}} -> parse columns 1, 3 as date and call
          result 'foo'

        .. versionadded:: 1.5.0

    encoding : str, optional, default 'utf-8'
        Encoding of XML document.

    parser : {{'lxml','etree'}}, default 'lxml'
        Parser module to use for retrieval of data. Only 'lxml' and
        'etree' are supported. With 'lxml' more complex XPath searches
        and ability to use XSLT stylesheet are supported.

    stylesheet : str, path object or file-like object
        A URL, file-like object, or a raw string containing an XSLT script.
        This stylesheet should flatten complex, deeply nested XML documents
        for easier parsing. To use this feature you must have ``lxml`` module
        installed and specify 'lxml' as ``parser``. The ``xpath`` must
        reference nodes of transformed XML document generated after XSLT
        transformation and not the original XML document. Only XSLT 1.0
        scripts and not later versions is currently supported.

    iterparse : dict, optional
        The nodes or attributes to retrieve in iterparsing of XML document
        as a dict with key being the name of repeating element and value being
        list of elements or attribute names that are descendants of the repeated
        element. Note: If this option is used, it will replace ``xpath`` parsing
        and unlike xpath, descendants do not need to relate to each other but can
        exist any where in document under the repeating element. This memory-
        efficient method should be used for very large XML files (500MB, 1GB, or 5GB+).
        For example, ::

            iterparse = {{"row_element": ["child_elem", "attr", "grandchild_elem"]}}

        .. versionadded:: 1.5.0

    {decompression_options}

        .. versionchanged:: 1.4.0 Zstandard support.

    {storage_options}

    Returns
    -------
    df
        A DataFrame.

    See Also
    --------
    read_json : Convert a JSON string to pandas object.
    read_html : Read HTML tables into a list of DataFrame objects.

    Notes
    -----
    This method is best designed to import shallow XML documents in
    following format which is the ideal fit for the two-dimensions of a
    ``DataFrame`` (row by column). ::

            <root>
                <row>
                  <column1>data</column1>
                  <column2>data</column2>
                  <column3>data</column3>
                  ...
               </row>
               <row>
                  ...
               </row>
               ...
            </root>

    As a file format, XML documents can be designed any way including
    layout of elements and attributes as long as it conforms to W3C
    specifications. Therefore, this method is a convenience handler for
    a specific flatter design and not all possible XML structures.

    However, for more complex XML documents, ``stylesheet`` allows you to
    temporarily redesign original document with XSLT (a special purpose
    language) for a flatter version for migration to a DataFrame.

    This function will *always* return a single :class:`DataFrame` or raise
    exceptions due to issues with XML document, ``xpath``, or other
    parameters.

    See the :ref:`read_xml documentation in the IO section of the docs
    <io.read_xml>` for more information in using this method to parse XML
    files to DataFrames.

    Examples
    --------
    >>> xml = '''<?xml version='1.0' encoding='utf-8'?>
    ... <data xmlns="http://example.com">
    ...  <row>
    ...    <shape>square</shape>
    ...    <degrees>360</degrees>
    ...    <sides>4.0</sides>
    ...  </row>
    ...  <row>
    ...    <shape>circle</shape>
    ...    <degrees>360</degrees>
    ...    <sides/>
    ...  </row>
    ...  <row>
    ...    <shape>triangle</shape>
    ...    <degrees>180</degrees>
    ...    <sides>3.0</sides>
    ...  </row>
    ... </data>'''

    >>> df = pd.read_xml(xml)
    >>> df
          shape  degrees  sides
    0    square      360    4.0
    1    circle      360    NaN
    2  triangle      180    3.0

    >>> xml = '''<?xml version='1.0' encoding='utf-8'?>
    ... <data>
    ...   <row shape="square" degrees="360" sides="4.0"/>
    ...   <row shape="circle" degrees="360"/>
    ...   <row shape="triangle" degrees="180" sides="3.0"/>
    ... </data>'''

    >>> df = pd.read_xml(xml, xpath=".//row")
    >>> df
          shape  degrees  sides
    0    square      360    4.0
    1    circle      360    NaN
    2  triangle      180    3.0

    >>> xml = '''<?xml version='1.0' encoding='utf-8'?>
    ... <doc:data xmlns:doc="https://example.com">
    ...   <doc:row>
    ...     <doc:shape>square</doc:shape>
    ...     <doc:degrees>360</doc:degrees>
    ...     <doc:sides>4.0</doc:sides>
    ...   </doc:row>
    ...   <doc:row>
    ...     <doc:shape>circle</doc:shape>
    ...     <doc:degrees>360</doc:degrees>
    ...     <doc:sides/>
    ...   </doc:row>
    ...   <doc:row>
    ...     <doc:shape>triangle</doc:shape>
    ...     <doc:degrees>180</doc:degrees>
    ...     <doc:sides>3.0</doc:sides>
    ...   </doc:row>
    ... </doc:data>'''

    >>> df = pd.read_xml(xml,
    ...                  xpath="//doc:row",
    ...                  namespaces={{"doc": "https://example.com"}})
    >>> df
          shape  degrees  sides
    0    square      360    4.0
    1    circle      360    NaN
    2  triangle      180    3.0
    """
    ...




"""
This type stub file was generated by pyright.
"""

from typing import Iterable, Literal, Mapping, TYPE_CHECKING, overload
from pandas._typing import Axis, HashableT
from pandas.util._decorators import deprecate_nonkeyword_arguments
from pandas import DataFrame, Series
from pandas.core.generic import NDFrame

"""
Concat routines.
"""
if TYPE_CHECKING:
    ...
@overload
def concat(objs: Iterable[DataFrame] | Mapping[HashableT, DataFrame], axis: Literal[0, "index"] = ..., join: str = ..., ignore_index: bool = ..., keys=..., levels=..., names=..., verify_integrity: bool = ..., sort: bool = ..., copy: bool = ...) -> DataFrame:
    ...

@overload
def concat(objs: Iterable[Series] | Mapping[HashableT, Series], axis: Literal[0, "index"] = ..., join: str = ..., ignore_index: bool = ..., keys=..., levels=..., names=..., verify_integrity: bool = ..., sort: bool = ..., copy: bool = ...) -> Series:
    ...

@overload
def concat(objs: Iterable[NDFrame] | Mapping[HashableT, NDFrame], axis: Literal[0, "index"] = ..., join: str = ..., ignore_index: bool = ..., keys=..., levels=..., names=..., verify_integrity: bool = ..., sort: bool = ..., copy: bool = ...) -> DataFrame | Series:
    ...

@overload
def concat(objs: Iterable[NDFrame] | Mapping[HashableT, NDFrame], axis: Literal[1, "columns"], join: str = ..., ignore_index: bool = ..., keys=..., levels=..., names=..., verify_integrity: bool = ..., sort: bool = ..., copy: bool = ...) -> DataFrame:
    ...

@overload
def concat(objs: Iterable[NDFrame] | Mapping[HashableT, NDFrame], axis: Axis = ..., join: str = ..., ignore_index: bool = ..., keys=..., levels=..., names=..., verify_integrity: bool = ..., sort: bool = ..., copy: bool = ...) -> DataFrame | Series:
    ...

@deprecate_nonkeyword_arguments(version=None, allowed_args=["objs"])
def concat(objs: Iterable[NDFrame] | Mapping[HashableT, NDFrame], axis: Axis = ..., join: str = ..., ignore_index: bool = ..., keys=..., levels=..., names=..., verify_integrity: bool = ..., sort: bool = ..., copy: bool = ...) -> DataFrame | Series:
    """
    Concatenate pandas objects along a particular axis.

    Allows optional set logic along the other axes.

    Can also add a layer of hierarchical indexing on the concatenation axis,
    which may be useful if the labels are the same (or overlapping) on
    the passed axis number.

    Parameters
    ----------
    objs : a sequence or mapping of Series or DataFrame objects
        If a mapping is passed, the sorted keys will be used as the `keys`
        argument, unless it is passed, in which case the values will be
        selected (see below). Any None objects will be dropped silently unless
        they are all None in which case a ValueError will be raised.
    axis : {0/'index', 1/'columns'}, default 0
        The axis to concatenate along.
    join : {'inner', 'outer'}, default 'outer'
        How to handle indexes on other axis (or axes).
    ignore_index : bool, default False
        If True, do not use the index values along the concatenation axis. The
        resulting axis will be labeled 0, ..., n - 1. This is useful if you are
        concatenating objects where the concatenation axis does not have
        meaningful indexing information. Note the index values on the other
        axes are still respected in the join.
    keys : sequence, default None
        If multiple levels passed, should contain tuples. Construct
        hierarchical index using the passed keys as the outermost level.
    levels : list of sequences, default None
        Specific levels (unique values) to use for constructing a
        MultiIndex. Otherwise they will be inferred from the keys.
    names : list, default None
        Names for the levels in the resulting hierarchical index.
    verify_integrity : bool, default False
        Check whether the new concatenated axis contains duplicates. This can
        be very expensive relative to the actual data concatenation.
    sort : bool, default False
        Sort non-concatenation axis if it is not already aligned when `join`
        is 'outer'.
        This has no effect when ``join='inner'``, which already preserves
        the order of the non-concatenation axis.

        .. versionchanged:: 1.0.0

           Changed to not sort by default.

    copy : bool, default True
        If False, do not copy data unnecessarily.

    Returns
    -------
    object, type of objs
        When concatenating all ``Series`` along the index (axis=0), a
        ``Series`` is returned. When ``objs`` contains at least one
        ``DataFrame``, a ``DataFrame`` is returned. When concatenating along
        the columns (axis=1), a ``DataFrame`` is returned.

    See Also
    --------
    DataFrame.join : Join DataFrames using indexes.
    DataFrame.merge : Merge DataFrames by indexes or columns.

    Notes
    -----
    The keys, levels, and names arguments are all optional.

    A walkthrough of how this method fits in with other tools for combining
    pandas objects can be found `here
    <https://pandas.pydata.org/pandas-docs/stable/user_guide/merging.html>`__.

    It is not recommended to build DataFrames by adding single rows in a
    for loop. Build a list of rows and make a DataFrame in a single concat.

    Examples
    --------
    Combine two ``Series``.

    >>> s1 = pd.Series(['a', 'b'])
    >>> s2 = pd.Series(['c', 'd'])
    >>> pd.concat([s1, s2])
    0    a
    1    b
    0    c
    1    d
    dtype: object

    Clear the existing index and reset it in the result
    by setting the ``ignore_index`` option to ``True``.

    >>> pd.concat([s1, s2], ignore_index=True)
    0    a
    1    b
    2    c
    3    d
    dtype: object

    Add a hierarchical index at the outermost level of
    the data with the ``keys`` option.

    >>> pd.concat([s1, s2], keys=['s1', 's2'])
    s1  0    a
        1    b
    s2  0    c
        1    d
    dtype: object

    Label the index keys you create with the ``names`` option.

    >>> pd.concat([s1, s2], keys=['s1', 's2'],
    ...           names=['Series name', 'Row ID'])
    Series name  Row ID
    s1           0         a
                 1         b
    s2           0         c
                 1         d
    dtype: object

    Combine two ``DataFrame`` objects with identical columns.

    >>> df1 = pd.DataFrame([['a', 1], ['b', 2]],
    ...                    columns=['letter', 'number'])
    >>> df1
      letter  number
    0      a       1
    1      b       2
    >>> df2 = pd.DataFrame([['c', 3], ['d', 4]],
    ...                    columns=['letter', 'number'])
    >>> df2
      letter  number
    0      c       3
    1      d       4
    >>> pd.concat([df1, df2])
      letter  number
    0      a       1
    1      b       2
    0      c       3
    1      d       4

    Combine ``DataFrame`` objects with overlapping columns
    and return everything. Columns outside the intersection will
    be filled with ``NaN`` values.

    >>> df3 = pd.DataFrame([['c', 3, 'cat'], ['d', 4, 'dog']],
    ...                    columns=['letter', 'number', 'animal'])
    >>> df3
      letter  number animal
    0      c       3    cat
    1      d       4    dog
    >>> pd.concat([df1, df3], sort=False)
      letter  number animal
    0      a       1    NaN
    1      b       2    NaN
    0      c       3    cat
    1      d       4    dog

    Combine ``DataFrame`` objects with overlapping columns
    and return only those that are shared by passing ``inner`` to
    the ``join`` keyword argument.

    >>> pd.concat([df1, df3], join="inner")
      letter  number
    0      a       1
    1      b       2
    0      c       3
    1      d       4

    Combine ``DataFrame`` objects horizontally along the x axis by
    passing in ``axis=1``.

    >>> df4 = pd.DataFrame([['bird', 'polly'], ['monkey', 'george']],
    ...                    columns=['animal', 'name'])
    >>> pd.concat([df1, df4], axis=1)
      letter  number  animal    name
    0      a       1    bird   polly
    1      b       2  monkey  george

    Prevent the result from including duplicate index values with the
    ``verify_integrity`` option.

    >>> df5 = pd.DataFrame([1], index=['a'])
    >>> df5
       0
    a  1
    >>> df6 = pd.DataFrame([2], index=['a'])
    >>> df6
       0
    a  2
    >>> pd.concat([df5, df6], verify_integrity=True)
    Traceback (most recent call last):
        ...
    ValueError: Indexes have overlapping values: ['a']

    Append a single row to the end of a ``DataFrame`` object.

    >>> df7 = pd.DataFrame({'a': 1, 'b': 2}, index=[0])
    >>> df7
        a   b
    0   1   2
    >>> new_row = pd.Series({'a': 3, 'b': 4})
    >>> new_row
    a    3
    b    4
    dtype: int64
    >>> pd.concat([df7, new_row.to_frame().T], ignore_index=True)
        a   b
    0   1   2
    1   3   4
    """
    ...

class _Concatenator:
    """
    Orchestrates a concatenation operation for BlockManagers
    """
    def __init__(self, objs: Iterable[NDFrame] | Mapping[HashableT, NDFrame], axis=..., join: str = ..., keys=..., levels=..., names=..., ignore_index: bool = ..., verify_integrity: bool = ..., copy: bool = ..., sort=...) -> None:
        ...
    
    def get_result(self): # -> Series | DataFrame:
        ...
    





"""
This type stub file was generated by pyright.
"""

from abc import ABC, abstractmethod
from typing import Iterable, Mapping, Sequence, TYPE_CHECKING
from pandas._typing import Dtype, WriteBuffer
from pandas import DataFrame, Index, Series

if TYPE_CHECKING:
    ...
frame_max_cols_sub = ...
show_counts_sub = ...
null_counts_sub = ...
frame_examples_sub = ...
frame_see_also_sub = ...
frame_sub_kwargs = ...
series_examples_sub = ...
series_see_also_sub = ...
series_sub_kwargs = ...
INFO_DOCSTRING = ...
class BaseInfo(ABC):
    """
    Base class for DataFrameInfo and SeriesInfo.

    Parameters
    ----------
    data : DataFrame or Series
        Either dataframe or series.
    memory_usage : bool or str, optional
        If "deep", introspect the data deeply by interrogating object dtypes
        for system-level memory consumption, and include it in the returned
        values.
    """
    data: DataFrame | Series
    memory_usage: bool | str
    @property
    @abstractmethod
    def dtypes(self) -> Iterable[Dtype]:
        """
        Dtypes.

        Returns
        -------
        dtypes : sequence
            Dtype of each of the DataFrame's columns (or one series column).
        """
        ...
    
    @property
    @abstractmethod
    def dtype_counts(self) -> Mapping[str, int]:
        """Mapping dtype - number of counts."""
        ...
    
    @property
    @abstractmethod
    def non_null_counts(self) -> Sequence[int]:
        """Sequence of non-null counts for all columns or column (if series)."""
        ...
    
    @property
    @abstractmethod
    def memory_usage_bytes(self) -> int:
        """
        Memory usage in bytes.

        Returns
        -------
        memory_usage_bytes : int
            Object's total memory usage in bytes.
        """
        ...
    
    @property
    def memory_usage_string(self) -> str:
        """Memory usage in a form of human readable string."""
        ...
    
    @property
    def size_qualifier(self) -> str:
        ...
    
    @abstractmethod
    def render(self, *, buf: WriteBuffer[str] | None, max_cols: int | None, verbose: bool | None, show_counts: bool | None) -> None:
        ...
    


class DataFrameInfo(BaseInfo):
    """
    Class storing dataframe-specific info.
    """
    def __init__(self, data: DataFrame, memory_usage: bool | str | None = ...) -> None:
        ...
    
    @property
    def dtype_counts(self) -> Mapping[str, int]:
        ...
    
    @property
    def dtypes(self) -> Iterable[Dtype]:
        """
        Dtypes.

        Returns
        -------
        dtypes
            Dtype of each of the DataFrame's columns.
        """
        ...
    
    @property
    def ids(self) -> Index:
        """
        Column names.

        Returns
        -------
        ids : Index
            DataFrame's column names.
        """
        ...
    
    @property
    def col_count(self) -> int:
        """Number of columns to be summarized."""
        ...
    
    @property
    def non_null_counts(self) -> Sequence[int]:
        """Sequence of non-null counts for all columns or column (if series)."""
        ...
    
    @property
    def memory_usage_bytes(self) -> int:
        ...
    
    def render(self, *, buf: WriteBuffer[str] | None, max_cols: int | None, verbose: bool | None, show_counts: bool | None) -> None:
        ...
    


class SeriesInfo(BaseInfo):
    """
    Class storing series-specific info.
    """
    def __init__(self, data: Series, memory_usage: bool | str | None = ...) -> None:
        ...
    
    def render(self, *, buf: WriteBuffer[str] | None = ..., max_cols: int | None = ..., verbose: bool | None = ..., show_counts: bool | None = ...) -> None:
        ...
    
    @property
    def non_null_counts(self) -> Sequence[int]:
        ...
    
    @property
    def dtypes(self) -> Iterable[Dtype]:
        ...
    
    @property
    def dtype_counts(self) -> Mapping[str, int]:
        ...
    
    @property
    def memory_usage_bytes(self) -> int:
        """Memory usage in bytes.

        Returns
        -------
        memory_usage_bytes : int
            Object's total memory usage in bytes.
        """
        ...
    


class InfoPrinterAbstract:
    """
    Class for printing dataframe or series info.
    """
    def to_buffer(self, buf: WriteBuffer[str] | None = ...) -> None:
        """Save dataframe info into buffer."""
        ...
    


class DataFrameInfoPrinter(InfoPrinterAbstract):
    """
    Class for printing dataframe info.

    Parameters
    ----------
    info : DataFrameInfo
        Instance of DataFrameInfo.
    max_cols : int, optional
        When to switch from the verbose to the truncated output.
    verbose : bool, optional
        Whether to print the full summary.
    show_counts : bool, optional
        Whether to show the non-null counts.
    """
    def __init__(self, info: DataFrameInfo, max_cols: int | None = ..., verbose: bool | None = ..., show_counts: bool | None = ...) -> None:
        ...
    
    @property
    def max_rows(self) -> int:
        """Maximum info rows to be displayed."""
        ...
    
    @property
    def exceeds_info_cols(self) -> bool:
        """Check if number of columns to be summarized does not exceed maximum."""
        ...
    
    @property
    def exceeds_info_rows(self) -> bool:
        """Check if number of rows to be summarized does not exceed maximum."""
        ...
    
    @property
    def col_count(self) -> int:
        """Number of columns to be summarized."""
        ...
    


class SeriesInfoPrinter(InfoPrinterAbstract):
    """Class for printing series info.

    Parameters
    ----------
    info : SeriesInfo
        Instance of SeriesInfo.
    verbose : bool, optional
        Whether to print the full summary.
    show_counts : bool, optional
        Whether to show the non-null counts.
    """
    def __init__(self, info: SeriesInfo, verbose: bool | None = ..., show_counts: bool | None = ...) -> None:
        ...
    


class TableBuilderAbstract(ABC):
    """
    Abstract builder for info table.
    """
    _lines: list[str]
    info: BaseInfo
    @abstractmethod
    def get_lines(self) -> list[str]:
        """Product in a form of list of lines (strings)."""
        ...
    
    @property
    def data(self) -> DataFrame | Series:
        ...
    
    @property
    def dtypes(self) -> Iterable[Dtype]:
        """Dtypes of each of the DataFrame's columns."""
        ...
    
    @property
    def dtype_counts(self) -> Mapping[str, int]:
        """Mapping dtype - number of counts."""
        ...
    
    @property
    def display_memory_usage(self) -> bool:
        """Whether to display memory usage."""
        ...
    
    @property
    def memory_usage_string(self) -> str:
        """Memory usage string with proper size qualifier."""
        ...
    
    @property
    def non_null_counts(self) -> Sequence[int]:
        ...
    
    def add_object_type_line(self) -> None:
        """Add line with string representation of dataframe to the table."""
        ...
    
    def add_index_range_line(self) -> None:
        """Add line with range of indices to the table."""
        ...
    
    def add_dtypes_line(self) -> None:
        """Add summary line with dtypes present in dataframe."""
        ...
    


class DataFrameTableBuilder(TableBuilderAbstract):
    """
    Abstract builder for dataframe info table.

    Parameters
    ----------
    info : DataFrameInfo.
        Instance of DataFrameInfo.
    """
    def __init__(self, *, info: DataFrameInfo) -> None:
        ...
    
    def get_lines(self) -> list[str]:
        ...
    
    @property
    def data(self) -> DataFrame:
        """DataFrame."""
        ...
    
    @property
    def ids(self) -> Index:
        """Dataframe columns."""
        ...
    
    @property
    def col_count(self) -> int:
        """Number of dataframe columns to be summarized."""
        ...
    
    def add_memory_usage_line(self) -> None:
        """Add line containing memory usage."""
        ...
    


class DataFrameTableBuilderNonVerbose(DataFrameTableBuilder):
    """
    Dataframe info table builder for non-verbose output.
    """
    def add_columns_summary_line(self) -> None:
        ...
    


class TableBuilderVerboseMixin(TableBuilderAbstract):
    """
    Mixin for verbose info output.
    """
    SPACING: str = ...
    strrows: Sequence[Sequence[str]]
    gross_column_widths: Sequence[int]
    with_counts: bool
    @property
    @abstractmethod
    def headers(self) -> Sequence[str]:
        """Headers names of the columns in verbose table."""
        ...
    
    @property
    def header_column_widths(self) -> Sequence[int]:
        """Widths of header columns (only titles)."""
        ...
    
    def add_header_line(self) -> None:
        ...
    
    def add_separator_line(self) -> None:
        ...
    
    def add_body_lines(self) -> None:
        ...
    


class DataFrameTableBuilderVerbose(DataFrameTableBuilder, TableBuilderVerboseMixin):
    """
    Dataframe info table builder for verbose output.
    """
    def __init__(self, *, info: DataFrameInfo, with_counts: bool) -> None:
        ...
    
    @property
    def headers(self) -> Sequence[str]:
        """Headers names of the columns in verbose table."""
        ...
    
    def add_columns_summary_line(self) -> None:
        ...
    


class SeriesTableBuilder(TableBuilderAbstract):
    """
    Abstract builder for series info table.

    Parameters
    ----------
    info : SeriesInfo.
        Instance of SeriesInfo.
    """
    def __init__(self, *, info: SeriesInfo) -> None:
        ...
    
    def get_lines(self) -> list[str]:
        ...
    
    @property
    def data(self) -> Series:
        """Series."""
        ...
    
    def add_memory_usage_line(self) -> None:
        """Add line containing memory usage."""
        ...
    


class SeriesTableBuilderNonVerbose(SeriesTableBuilder):
    """
    Series info table builder for non-verbose output.
    """
    ...


class SeriesTableBuilderVerbose(SeriesTableBuilder, TableBuilderVerboseMixin):
    """
    Series info table builder for verbose output.
    """
    def __init__(self, *, info: SeriesInfo, with_counts: bool) -> None:
        ...
    
    def add_series_name_line(self) -> None:
        ...
    
    @property
    def headers(self) -> Sequence[str]:
        """Headers names of the columns in verbose table."""
        ...
    





"""
This type stub file was generated by pyright.
"""

import numpy as np
from datetime import tzinfo
from pandas._libs.tslibs.dtypes import Resolution
from pandas._libs.tslibs.offsets import BaseOffset
from pandas._typing import npt

"""
For cython types that cannot be represented precisely, closest-available
python equivalents are used, and the precise types kept as adjacent comments.
"""
def dt64arr_to_periodarr(stamps: npt.NDArray[np.int64], freq: int, tz: tzinfo | None, reso: int = ...) -> npt.NDArray[np.int64]:
    ...

def is_date_array_normalized(stamps: npt.NDArray[np.int64], tz: tzinfo | None, reso: int) -> bool:
    ...

def normalize_i8_timestamps(stamps: npt.NDArray[np.int64], tz: tzinfo | None, reso: int) -> npt.NDArray[np.int64]:
    ...

def get_resolution(stamps: npt.NDArray[np.int64], tz: tzinfo | None = ..., reso: int = ...) -> Resolution:
    ...

def ints_to_pydatetime(arr: npt.NDArray[np.int64], tz: tzinfo | None = ..., freq: BaseOffset | None = ..., fold: bool = ..., box: str = ..., reso: int = ...) -> npt.NDArray[np.object_]:
    ...

def tz_convert_from_utc(stamps: npt.NDArray[np.int64], tz: tzinfo | None, reso: int = ...) -> npt.NDArray[np.int64]:
    ...




"""
This type stub file was generated by pyright.
"""

import numpy as np
from typing import Literal
from pandas._typing import npt

def group_median_float64(out: np.ndarray, counts: npt.NDArray[np.int64], values: np.ndarray, labels: npt.NDArray[np.int64], min_count: int = ...) -> None:
    ...

def group_cumprod_float64(out: np.ndarray, values: np.ndarray, labels: np.ndarray, ngroups: int, is_datetimelike: bool, skipna: bool = ...) -> None:
    ...

def group_cumsum(out: np.ndarray, values: np.ndarray, labels: np.ndarray, ngroups: int, is_datetimelike: bool, skipna: bool = ..., mask: np.ndarray | None = ..., result_mask: np.ndarray | None = ...) -> None:
    ...

def group_shift_indexer(out: np.ndarray, labels: np.ndarray, ngroups: int, periods: int) -> None:
    ...

def group_fillna_indexer(out: np.ndarray, labels: np.ndarray, sorted_labels: npt.NDArray[np.intp], mask: npt.NDArray[np.uint8], direction: Literal["ffill", "bfill"], limit: int, dropna: bool) -> None:
    ...

def group_any_all(out: np.ndarray, values: np.ndarray, labels: np.ndarray, mask: np.ndarray, val_test: Literal["any", "all"], skipna: bool) -> None:
    ...

def group_sum(out: np.ndarray, counts: np.ndarray, values: np.ndarray, labels: np.ndarray, mask: np.ndarray | None, result_mask: np.ndarray | None = ..., min_count: int = ..., is_datetimelike: bool = ...) -> None:
    ...

def group_prod(out: np.ndarray, counts: np.ndarray, values: np.ndarray, labels: np.ndarray, mask: np.ndarray | None, result_mask: np.ndarray | None = ..., min_count: int = ...) -> None:
    ...

def group_var(out: np.ndarray, counts: np.ndarray, values: np.ndarray, labels: np.ndarray, min_count: int = ..., ddof: int = ...) -> None:
    ...

def group_mean(out: np.ndarray, counts: np.ndarray, values: np.ndarray, labels: np.ndarray, min_count: int = ..., is_datetimelike: bool = ..., mask: np.ndarray | None = ..., result_mask: np.ndarray | None = ...) -> None:
    ...

def group_ohlc(out: np.ndarray, counts: np.ndarray, values: np.ndarray, labels: np.ndarray, min_count: int = ..., mask: np.ndarray | None = ..., result_mask: np.ndarray | None = ...) -> None:
    ...

def group_quantile(out: npt.NDArray[np.float64], values: np.ndarray, labels: npt.NDArray[np.intp], mask: npt.NDArray[np.uint8], sort_indexer: npt.NDArray[np.intp], qs: npt.NDArray[np.float64], interpolation: Literal["linear", "lower", "higher", "nearest", "midpoint"]) -> None:
    ...

def group_last(out: np.ndarray, counts: np.ndarray, values: np.ndarray, labels: np.ndarray, mask: npt.NDArray[np.bool_] | None, result_mask: npt.NDArray[np.bool_] | None = ..., min_count: int = ..., is_datetimelike: bool = ...) -> None:
    ...

def group_nth(out: np.ndarray, counts: np.ndarray, values: np.ndarray, labels: np.ndarray, mask: npt.NDArray[np.bool_] | None, result_mask: npt.NDArray[np.bool_] | None = ..., min_count: int = ..., rank: int = ..., is_datetimelike: bool = ...) -> None:
    ...

def group_rank(out: np.ndarray, values: np.ndarray, labels: np.ndarray, ngroups: int, is_datetimelike: bool, ties_method: Literal["average", "min", "max", "first", "dense"] = ..., ascending: bool = ..., pct: bool = ..., na_option: Literal["keep", "top", "bottom"] = ..., mask: npt.NDArray[np.bool_] | None = ...) -> None:
    ...

def group_max(out: np.ndarray, counts: np.ndarray, values: np.ndarray, labels: np.ndarray, min_count: int = ..., is_datetimelike: bool = ..., mask: np.ndarray | None = ..., result_mask: np.ndarray | None = ...) -> None:
    ...

def group_min(out: np.ndarray, counts: np.ndarray, values: np.ndarray, labels: np.ndarray, min_count: int = ..., is_datetimelike: bool = ..., mask: np.ndarray | None = ..., result_mask: np.ndarray | None = ...) -> None:
    ...

def group_cummin(out: np.ndarray, values: np.ndarray, labels: np.ndarray, ngroups: int, is_datetimelike: bool, mask: np.ndarray | None = ..., result_mask: np.ndarray | None = ..., skipna: bool = ...) -> None:
    ...

def group_cummax(out: np.ndarray, values: np.ndarray, labels: np.ndarray, ngroups: int, is_datetimelike: bool, mask: np.ndarray | None = ..., result_mask: np.ndarray | None = ..., skipna: bool = ...) -> None:
    ...




"""
This type stub file was generated by pyright.
"""

import numpy as np
from pandas._typing import NumpyIndexT

def cartesian_product(X) -> list[np.ndarray]:
    """
    Numpy version of itertools.product.
    Sometimes faster (for large inputs)...

    Parameters
    ----------
    X : list-like of list-likes

    Returns
    -------
    product : list of ndarrays

    Examples
    --------
    >>> cartesian_product([list('ABC'), [1, 2]])
    [array(['A', 'A', 'B', 'B', 'C', 'C'], dtype='<U1'), array([1, 2, 1, 2, 1, 2])]

    See Also
    --------
    itertools.product : Cartesian product of input iterables.  Equivalent to
        nested for-loops.
    """
    ...

def tile_compat(arr: NumpyIndexT, num: int) -> NumpyIndexT:
    """
    Index compat for np.tile.

    Notes
    -----
    Does not support multi-dimensional `num`.
    """
    ...




"""
This type stub file was generated by pyright.
"""

import numpy as np
from textwrap import dedent
from typing import Literal, TYPE_CHECKING, final
from pandas._libs import lib
from pandas._typing import AnyArrayLike, ArrayLike, DtypeObj, IndexLabel, NumpySorter, NumpyValueArrayLike, TakeIndexer, npt
from pandas.util._decorators import doc
from pandas import DataFrame, Index, Series
from pandas.core.arrays import ExtensionArray

"""
Generic data algorithms. This module is experimental at the moment and not
intended for public consumption
"""
if TYPE_CHECKING:
    ...
_hashtables = ...
def unique(values): # -> ArrayLike | tuple[ArrayLike, NDArray[Any]]:
    """
    Return unique values based on a hash table.

    Uniques are returned in order of appearance. This does NOT sort.

    Significantly faster than numpy.unique for long enough sequences.
    Includes NA values.

    Parameters
    ----------
    values : 1d array-like

    Returns
    -------
    numpy.ndarray or ExtensionArray

        The return can be:

        * Index : when the input is an Index
        * Categorical : when the input is a Categorical dtype
        * ndarray : when the input is a Series/ndarray

        Return numpy.ndarray or ExtensionArray.

    See Also
    --------
    Index.unique : Return unique values from an Index.
    Series.unique : Return unique values of Series object.

    Examples
    --------
    >>> pd.unique(pd.Series([2, 1, 3, 3]))
    array([2, 1, 3])

    >>> pd.unique(pd.Series([2] + [1] * 5))
    array([2, 1])

    >>> pd.unique(pd.Series([pd.Timestamp("20160101"), pd.Timestamp("20160101")]))
    array(['2016-01-01T00:00:00.000000000'], dtype='datetime64[ns]')

    >>> pd.unique(
    ...     pd.Series(
    ...         [
    ...             pd.Timestamp("20160101", tz="US/Eastern"),
    ...             pd.Timestamp("20160101", tz="US/Eastern"),
    ...         ]
    ...     )
    ... )
    <DatetimeArray>
    ['2016-01-01 00:00:00-05:00']
    Length: 1, dtype: datetime64[ns, US/Eastern]

    >>> pd.unique(
    ...     pd.Index(
    ...         [
    ...             pd.Timestamp("20160101", tz="US/Eastern"),
    ...             pd.Timestamp("20160101", tz="US/Eastern"),
    ...         ]
    ...     )
    ... )
    DatetimeIndex(['2016-01-01 00:00:00-05:00'],
            dtype='datetime64[ns, US/Eastern]',
            freq=None)

    >>> pd.unique(list("baabc"))
    array(['b', 'a', 'c'], dtype=object)

    An unordered Categorical will return categories in the
    order of appearance.

    >>> pd.unique(pd.Series(pd.Categorical(list("baabc"))))
    ['b', 'a', 'c']
    Categories (3, object): ['a', 'b', 'c']

    >>> pd.unique(pd.Series(pd.Categorical(list("baabc"), categories=list("abc"))))
    ['b', 'a', 'c']
    Categories (3, object): ['a', 'b', 'c']

    An ordered Categorical preserves the category ordering.

    >>> pd.unique(
    ...     pd.Series(
    ...         pd.Categorical(list("baabc"), categories=list("abc"), ordered=True)
    ...     )
    ... )
    ['b', 'a', 'c']
    Categories (3, object): ['a' < 'b' < 'c']

    An array of tuples

    >>> pd.unique([("a", "b"), ("b", "a"), ("a", "c"), ("b", "a")])
    array([('a', 'b'), ('b', 'a'), ('a', 'c')], dtype=object)
    """
    ...

def unique_with_mask(values, mask: npt.NDArray[np.bool_] | None = ...): # -> ArrayLike | tuple[ArrayLike, NDArray[Any]]:
    """See algorithms.unique for docs. Takes a mask for masked arrays."""
    ...

unique1d = ...
def isin(comps: AnyArrayLike, values: AnyArrayLike) -> npt.NDArray[np.bool_]:
    """
    Compute the isin boolean array.

    Parameters
    ----------
    comps : array-like
    values : array-like

    Returns
    -------
    ndarray[bool]
        Same length as `comps`.
    """
    ...

def factorize_array(values: np.ndarray, na_sentinel: int | None = ..., size_hint: int | None = ..., na_value: object = ..., mask: npt.NDArray[np.bool_] | None = ...) -> tuple[npt.NDArray[np.intp], np.ndarray]:
    """
    Factorize a numpy array to codes and uniques.

    This doesn't do any coercion of types or unboxing before factorization.

    Parameters
    ----------
    values : ndarray
    na_sentinel : int, default -1
    size_hint : int, optional
        Passed through to the hashtable's 'get_labels' method
    na_value : object, optional
        A value in `values` to consider missing. Note: only use this
        parameter when you know that you don't have any values pandas would
        consider missing in the array (NaN for float data, iNaT for
        datetimes, etc.).
    mask : ndarray[bool], optional
        If not None, the mask is used as indicator for missing values
        (True = missing, False = valid) instead of `na_value` or
        condition "val != val".

    Returns
    -------
    codes : ndarray[np.intp]
    uniques : ndarray
    """
    ...

@doc(values=dedent("""\
    values : sequence
        A 1-D sequence. Sequences that aren't pandas objects are
        coerced to ndarrays before factorization.
    """), sort=dedent("""\
    sort : bool, default False
        Sort `uniques` and shuffle `codes` to maintain the
        relationship.
    """), size_hint=dedent("""\
    size_hint : int, optional
        Hint to the hashtable sizer.
    """))
def factorize(values, sort: bool = ..., na_sentinel: int | None | lib.NoDefault = ..., use_na_sentinel: bool | lib.NoDefault = ..., size_hint: int | None = ...) -> tuple[np.ndarray, np.ndarray | Index]:
    """
    Encode the object as an enumerated type or categorical variable.

    This method is useful for obtaining a numeric representation of an
    array when all that matters is identifying distinct values. `factorize`
    is available as both a top-level function :func:`pandas.factorize`,
    and as a method :meth:`Series.factorize` and :meth:`Index.factorize`.

    Parameters
    ----------
    {values}{sort}
    na_sentinel : int or None, default -1
        Value to mark "not found". If None, will not drop the NaN
        from the uniques of the values.

        .. deprecated:: 1.5.0
            The na_sentinel argument is deprecated and
            will be removed in a future version of pandas. Specify use_na_sentinel as
            either True or False.

        .. versionchanged:: 1.1.2

    use_na_sentinel : bool, default True
        If True, the sentinel -1 will be used for NaN values. If False,
        NaN values will be encoded as non-negative integers and will not drop the
        NaN from the uniques of the values.

        .. versionadded:: 1.5.0
    {size_hint}\

    Returns
    -------
    codes : ndarray
        An integer ndarray that's an indexer into `uniques`.
        ``uniques.take(codes)`` will have the same values as `values`.
    uniques : ndarray, Index, or Categorical
        The unique valid values. When `values` is Categorical, `uniques`
        is a Categorical. When `values` is some other pandas object, an
        `Index` is returned. Otherwise, a 1-D ndarray is returned.

        .. note::

           Even if there's a missing value in `values`, `uniques` will
           *not* contain an entry for it.

    See Also
    --------
    cut : Discretize continuous-valued array.
    unique : Find the unique value in an array.

    Notes
    -----
    Reference :ref:`the user guide <reshaping.factorize>` for more examples.

    Examples
    --------
    These examples all show factorize as a top-level method like
    ``pd.factorize(values)``. The results are identical for methods like
    :meth:`Series.factorize`.

    >>> codes, uniques = pd.factorize(['b', 'b', 'a', 'c', 'b'])
    >>> codes
    array([0, 0, 1, 2, 0]...)
    >>> uniques
    array(['b', 'a', 'c'], dtype=object)

    With ``sort=True``, the `uniques` will be sorted, and `codes` will be
    shuffled so that the relationship is the maintained.

    >>> codes, uniques = pd.factorize(['b', 'b', 'a', 'c', 'b'], sort=True)
    >>> codes
    array([1, 1, 0, 2, 1]...)
    >>> uniques
    array(['a', 'b', 'c'], dtype=object)

    When ``use_na_sentinel=True`` (the default), missing values are indicated in
    the `codes` with the sentinel value ``-1`` and missing values are not
    included in `uniques`.

    >>> codes, uniques = pd.factorize(['b', None, 'a', 'c', 'b'])
    >>> codes
    array([ 0, -1,  1,  2,  0]...)
    >>> uniques
    array(['b', 'a', 'c'], dtype=object)

    Thus far, we've only factorized lists (which are internally coerced to
    NumPy arrays). When factorizing pandas objects, the type of `uniques`
    will differ. For Categoricals, a `Categorical` is returned.

    >>> cat = pd.Categorical(['a', 'a', 'c'], categories=['a', 'b', 'c'])
    >>> codes, uniques = pd.factorize(cat)
    >>> codes
    array([0, 0, 1]...)
    >>> uniques
    ['a', 'c']
    Categories (3, object): ['a', 'b', 'c']

    Notice that ``'b'`` is in ``uniques.categories``, despite not being
    present in ``cat.values``.

    For all other pandas objects, an Index of the appropriate type is
    returned.

    >>> cat = pd.Series(['a', 'a', 'c'])
    >>> codes, uniques = pd.factorize(cat)
    >>> codes
    array([0, 0, 1]...)
    >>> uniques
    Index(['a', 'c'], dtype='object')

    If NaN is in the values, and we want to include NaN in the uniques of the
    values, it can be achieved by setting ``use_na_sentinel=False``.

    >>> values = np.array([1, 2, 1, np.nan])
    >>> codes, uniques = pd.factorize(values)  # default: use_na_sentinel=True
    >>> codes
    array([ 0,  1,  0, -1])
    >>> uniques
    array([1., 2.])

    >>> codes, uniques = pd.factorize(values, use_na_sentinel=False)
    >>> codes
    array([0, 1, 0, 2])
    >>> uniques
    array([ 1.,  2., nan])
    """
    ...

def resolve_na_sentinel(na_sentinel: int | None | lib.NoDefault, use_na_sentinel: bool | lib.NoDefault) -> int | None:
    """
    Determine value of na_sentinel for factorize methods.

    See GH#46910 for details on the deprecation.

    Parameters
    ----------
    na_sentinel : int, None, or lib.no_default
        Value passed to the method.
    use_na_sentinel : bool or lib.no_default
        Value passed to the method.

    Returns
    -------
    Resolved value of na_sentinel.
    """
    ...

def value_counts(values, sort: bool = ..., ascending: bool = ..., normalize: bool = ..., bins=..., dropna: bool = ...) -> Series:
    """
    Compute a histogram of the counts of non-null values.

    Parameters
    ----------
    values : ndarray (1-d)
    sort : bool, default True
        Sort by values
    ascending : bool, default False
        Sort in ascending order
    normalize: bool, default False
        If True then compute a relative histogram
    bins : integer, optional
        Rather than count values, group them into half-open bins,
        convenience for pd.cut, only works with numeric data
    dropna : bool, default True
        Don't include counts of NaN

    Returns
    -------
    Series
    """
    ...

def value_counts_arraylike(values: np.ndarray, dropna: bool, mask: npt.NDArray[np.bool_] | None = ...) -> tuple[ArrayLike, npt.NDArray[np.int64]]:
    """
    Parameters
    ----------
    values : np.ndarray
    dropna : bool
    mask : np.ndarray[bool] or None, default None

    Returns
    -------
    uniques : np.ndarray
    counts : np.ndarray[np.int64]
    """
    ...

def duplicated(values: ArrayLike, keep: Literal["first", "last", False] = ...) -> npt.NDArray[np.bool_]:
    """
    Return boolean ndarray denoting duplicate values.

    Parameters
    ----------
    values : nd.array, ExtensionArray or Series
        Array over which to check for duplicate values.
    keep : {'first', 'last', False}, default 'first'
        - ``first`` : Mark duplicates as ``True`` except for the first
          occurrence.
        - ``last`` : Mark duplicates as ``True`` except for the last
          occurrence.
        - False : Mark all duplicates as ``True``.

    Returns
    -------
    duplicated : ndarray[bool]
    """
    ...

def mode(values: ArrayLike, dropna: bool = ..., mask: npt.NDArray[np.bool_] | None = ...) -> ArrayLike:
    """
    Returns the mode(s) of an array.

    Parameters
    ----------
    values : array-like
        Array over which to check for duplicate values.
    dropna : bool, default True
        Don't consider counts of NaN/NaT.

    Returns
    -------
    np.ndarray or ExtensionArray
    """
    ...

def rank(values: ArrayLike, axis: int = ..., method: str = ..., na_option: str = ..., ascending: bool = ..., pct: bool = ...) -> npt.NDArray[np.float64]:
    """
    Rank the values along a given axis.

    Parameters
    ----------
    values : np.ndarray or ExtensionArray
        Array whose values will be ranked. The number of dimensions in this
        array must not exceed 2.
    axis : int, default 0
        Axis over which to perform rankings.
    method : {'average', 'min', 'max', 'first', 'dense'}, default 'average'
        The method by which tiebreaks are broken during the ranking.
    na_option : {'keep', 'top'}, default 'keep'
        The method by which NaNs are placed in the ranking.
        - ``keep``: rank each NaN value with a NaN ranking
        - ``top``: replace each NaN with either +/- inf so that they
                   there are ranked at the top
    ascending : bool, default True
        Whether or not the elements should be ranked in ascending order.
    pct : bool, default False
        Whether or not to the display the returned rankings in integer form
        (e.g. 1, 2, 3) or in percentile form (e.g. 0.333..., 0.666..., 1).
    """
    ...

def checked_add_with_arr(arr: npt.NDArray[np.int64], b: int | npt.NDArray[np.int64], arr_mask: npt.NDArray[np.bool_] | None = ..., b_mask: npt.NDArray[np.bool_] | None = ...) -> npt.NDArray[np.int64]:
    """
    Perform array addition that checks for underflow and overflow.

    Performs the addition of an int64 array and an int64 integer (or array)
    but checks that they do not result in overflow first. For elements that
    are indicated to be NaN, whether or not there is overflow for that element
    is automatically ignored.

    Parameters
    ----------
    arr : np.ndarray[int64] addend.
    b : array or scalar addend.
    arr_mask : np.ndarray[bool] or None, default None
        array indicating which elements to exclude from checking
    b_mask : np.ndarray[bool] or None, default None
        array or scalar indicating which element(s) to exclude from checking

    Returns
    -------
    sum : An array for elements x + b for each element x in arr if b is
          a scalar or an array for elements x + y for each element pair
          (x, y) in (arr, b).

    Raises
    ------
    OverflowError if any x + y exceeds the maximum or minimum int64 value.
    """
    ...

class SelectN:
    def __init__(self, obj, n: int, keep: str) -> None:
        ...
    
    def compute(self, method: str) -> DataFrame | Series:
        ...
    
    @final
    def nlargest(self): # -> DataFrame | Series:
        ...
    
    @final
    def nsmallest(self): # -> DataFrame | Series:
        ...
    
    @final
    @staticmethod
    def is_valid_dtype_n_method(dtype: DtypeObj) -> bool:
        """
        Helper function to determine if dtype is valid for
        nsmallest/nlargest methods
        """
        ...
    


class SelectNSeries(SelectN):
    """
    Implement n largest/smallest for Series

    Parameters
    ----------
    obj : Series
    n : int
    keep : {'first', 'last'}, default 'first'

    Returns
    -------
    nordered : Series
    """
    def compute(self, method: str) -> Series:
        ...
    


class SelectNFrame(SelectN):
    """
    Implement n largest/smallest for DataFrame

    Parameters
    ----------
    obj : DataFrame
    n : int
    keep : {'first', 'last'}, default 'first'
    columns : list or str

    Returns
    -------
    nordered : DataFrame
    """
    def __init__(self, obj: DataFrame, n: int, keep: str, columns: IndexLabel) -> None:
        ...
    
    def compute(self, method: str) -> DataFrame:
        ...
    


def take(arr, indices: TakeIndexer, axis: int = ..., allow_fill: bool = ..., fill_value=...): # -> ndarray[Unknown, Unknown] | ndarray[Any, dtype[Unknown]]:
    """
    Take elements from an array.

    Parameters
    ----------
    arr : array-like or scalar value
        Non array-likes (sequences/scalars without a dtype) are coerced
        to an ndarray.
    indices : sequence of int or one-dimensional np.ndarray of int
        Indices to be taken.
    axis : int, default 0
        The axis over which to select values.
    allow_fill : bool, default False
        How to handle negative values in `indices`.

        * False: negative values in `indices` indicate positional indices
          from the right (the default). This is similar to :func:`numpy.take`.

        * True: negative values in `indices` indicate
          missing values. These values are set to `fill_value`. Any other
          negative values raise a ``ValueError``.

    fill_value : any, optional
        Fill value to use for NA-indices when `allow_fill` is True.
        This may be ``None``, in which case the default NA value for
        the type (``self.dtype.na_value``) is used.

        For multi-dimensional `arr`, each *element* is filled with
        `fill_value`.

    Returns
    -------
    ndarray or ExtensionArray
        Same type as the input.

    Raises
    ------
    IndexError
        When `indices` is out of bounds for the array.
    ValueError
        When the indexer contains negative values other than ``-1``
        and `allow_fill` is True.

    Notes
    -----
    When `allow_fill` is False, `indices` may be whatever dimensionality
    is accepted by NumPy for `arr`.

    When `allow_fill` is True, `indices` should be 1-D.

    See Also
    --------
    numpy.take : Take elements from an array along an axis.

    Examples
    --------
    >>> import pandas as pd

    With the default ``allow_fill=False``, negative numbers indicate
    positional indices from the right.

    >>> pd.api.extensions.take(np.array([10, 20, 30]), [0, 0, -1])
    array([10, 10, 30])

    Setting ``allow_fill=True`` will place `fill_value` in those positions.

    >>> pd.api.extensions.take(np.array([10, 20, 30]), [0, 0, -1], allow_fill=True)
    array([10., 10., nan])

    >>> pd.api.extensions.take(np.array([10, 20, 30]), [0, 0, -1], allow_fill=True,
    ...      fill_value=-10)
    array([ 10,  10, -10])
    """
    ...

def searchsorted(arr: ArrayLike, value: NumpyValueArrayLike | ExtensionArray, side: Literal["left", "right"] = ..., sorter: NumpySorter = ...) -> npt.NDArray[np.intp] | np.intp:
    """
    Find indices where elements should be inserted to maintain order.

    .. versionadded:: 0.25.0

    Find the indices into a sorted array `arr` (a) such that, if the
    corresponding elements in `value` were inserted before the indices,
    the order of `arr` would be preserved.

    Assuming that `arr` is sorted:

    ======  ================================
    `side`  returned index `i` satisfies
    ======  ================================
    left    ``arr[i-1] < value <= self[i]``
    right   ``arr[i-1] <= value < self[i]``
    ======  ================================

    Parameters
    ----------
    arr: np.ndarray, ExtensionArray, Series
        Input array. If `sorter` is None, then it must be sorted in
        ascending order, otherwise `sorter` must be an array of indices
        that sort it.
    value : array-like or scalar
        Values to insert into `arr`.
    side : {'left', 'right'}, optional
        If 'left', the index of the first suitable location found is given.
        If 'right', return the last such index.  If there is no suitable
        index, return either 0 or N (where N is the length of `self`).
    sorter : 1-D array-like, optional
        Optional array of integer indices that sort array a into ascending
        order. They are typically the result of argsort.

    Returns
    -------
    array of ints or int
        If value is array-like, array of insertion points.
        If value is scalar, a single integer.

    See Also
    --------
    numpy.searchsorted : Similar method from NumPy.
    """
    ...

_diff_special = ...
def diff(arr, n: int, axis: int = ...): # -> Any | ndarray[Any, dtype[Any]] | ndarray[Any, dtype[int64 | object_ | float32 | float64]]:
    """
    difference of n between self,
    analogous to s-s.shift(n)

    Parameters
    ----------
    arr : ndarray or ExtensionArray
    n : int
        number of periods
    axis : {0, 1}
        axis to shift on
    stacklevel : int, default 3
        The stacklevel for the lost dtype warning.

    Returns
    -------
    shifted
    """
    ...

def safe_sort(values, codes=..., na_sentinel: int = ..., assume_unique: bool = ..., verify: bool = ...) -> np.ndarray | tuple[np.ndarray, np.ndarray]:
    """
    Sort ``values`` and reorder corresponding ``codes``.

    ``values`` should be unique if ``codes`` is not None.
    Safe for use with mixed types (int, str), orders ints before strs.

    Parameters
    ----------
    values : list-like
        Sequence; must be unique if ``codes`` is not None.
    codes : list_like, optional
        Indices to ``values``. All out of bound indices are treated as
        "not found" and will be masked with ``na_sentinel``.
    na_sentinel : int, default -1
        Value in ``codes`` to mark "not found".
        Ignored when ``codes`` is None.
    assume_unique : bool, default False
        When True, ``values`` are assumed to be unique, which can speed up
        the calculation. Ignored when ``codes`` is None.
    verify : bool, default True
        Check if codes are out of bound for the values and put out of bound
        codes equal to na_sentinel. If ``verify=False``, it is assumed there
        are no out of bound codes. Ignored when ``codes`` is None.

        .. versionadded:: 0.25.0

    Returns
    -------
    ordered : ndarray
        Sorted ``values``
    new_codes : ndarray
        Reordered ``codes``; returned when ``codes`` is not None.

    Raises
    ------
    TypeError
        * If ``values`` is not list-like or if ``codes`` is neither None
        nor list-like
        * If ``values`` cannot be sorted
    ValueError
        * If ``codes`` is not None and ``values`` contain duplicates.
    """
    ...

def union_with_duplicates(lvals: ArrayLike, rvals: ArrayLike) -> ArrayLike:
    """
    Extracts the union from lvals and rvals with respect to duplicates and nans in
    both arrays.

    Parameters
    ----------
    lvals: np.ndarray or ExtensionArray
        left values which is ordered in front.
    rvals: np.ndarray or ExtensionArray
        right values ordered after lvals.

    Returns
    -------
    np.ndarray or ExtensionArray
        Containing the unsorted union of both arrays.

    Notes
    -----
    Caller is responsible for ensuring lvals.dtype == rvals.dtype.
    """
    ...




"""
This type stub file was generated by pyright.
"""

import numpy as np
from contextlib import contextmanager
from io import StringIO
from typing import Any, Callable, Final, Hashable, IO, Iterable, Iterator, Sequence, TYPE_CHECKING
from pandas._typing import ColspaceArgType, CompressionOptions, FilePath, FloatFormatType, FormattersType, IndexLabel, StorageOptions, WriteBuffer
from pandas.util._decorators import deprecate_kwarg
from pandas.core.arrays import Categorical, DatetimeArray, TimedeltaArray
from pandas.core.indexes.api import Index
from pandas.core.indexes.datetimes import DatetimeIndex
from pandas.core.indexes.timedeltas import TimedeltaIndex
from pandas import DataFrame, Series

"""
Internal module for formatting output data in csv, html, xml,
and latex files. This module also applies to display formatting.
"""
if TYPE_CHECKING:
    ...
common_docstring: Final = ...
_VALID_JUSTIFY_PARAMETERS = ...
return_docstring: Final = ...
class CategoricalFormatter:
    def __init__(self, categorical: Categorical, buf: IO[str] | None = ..., length: bool = ..., na_rep: str = ..., footer: bool = ...) -> None:
        ...
    
    def to_string(self) -> str:
        ...
    


class SeriesFormatter:
    def __init__(self, series: Series, buf: IO[str] | None = ..., length: bool | str = ..., header: bool = ..., index: bool = ..., na_rep: str = ..., name: bool = ..., float_format: str | None = ..., dtype: bool = ..., max_rows: int | None = ..., min_rows: int | None = ...) -> None:
        ...
    
    def to_string(self) -> str:
        ...
    


class TextAdjustment:
    def __init__(self) -> None:
        ...
    
    def len(self, text: str) -> int:
        ...
    
    def justify(self, texts: Any, max_len: int, mode: str = ...) -> list[str]:
        ...
    
    def adjoin(self, space: int, *lists, **kwargs) -> str:
        ...
    


class EastAsianTextAdjustment(TextAdjustment):
    def __init__(self) -> None:
        ...
    
    def len(self, text: str) -> int:
        """
        Calculate display width considering unicode East Asian Width
        """
        ...
    
    def justify(self, texts: Iterable[str], max_len: int, mode: str = ...) -> list[str]:
        ...
    


def get_adjustment() -> TextAdjustment:
    ...

def get_dataframe_repr_params() -> dict[str, Any]:
    """Get the parameters used to repr(dataFrame) calls using DataFrame.to_string.

    Supplying these parameters to DataFrame.to_string is equivalent to calling
    ``repr(DataFrame)``. This is useful if you want to adjust the repr output.

    .. versionadded:: 1.4.0

    Example
    -------
    >>> import pandas as pd
    >>>
    >>> df = pd.DataFrame([[1, 2], [3, 4]])
    >>> repr_params = pd.io.formats.format.get_dataframe_repr_params()
    >>> repr(df) == df.to_string(**repr_params)
    True
    """
    ...

def get_series_repr_params() -> dict[str, Any]:
    """Get the parameters used to repr(Series) calls using Series.to_string.

    Supplying these parameters to Series.to_string is equivalent to calling
    ``repr(series)``. This is useful if you want to adjust the series repr output.

    .. versionadded:: 1.4.0

    Example
    -------
    >>> import pandas as pd
    >>>
    >>> ser = pd.Series([1, 2, 3, 4])
    >>> repr_params = pd.io.formats.format.get_series_repr_params()
    >>> repr(ser) == ser.to_string(**repr_params)
    True
    """
    ...

class DataFrameFormatter:
    """Class for processing dataframe formatting options and data."""
    __doc__ = ...
    def __init__(self, frame: DataFrame, columns: Sequence[Hashable] | None = ..., col_space: ColspaceArgType | None = ..., header: bool | Sequence[str] = ..., index: bool = ..., na_rep: str = ..., formatters: FormattersType | None = ..., justify: str | None = ..., float_format: FloatFormatType | None = ..., sparsify: bool | None = ..., index_names: bool = ..., max_rows: int | None = ..., min_rows: int | None = ..., max_cols: int | None = ..., show_dimensions: bool | str = ..., decimal: str = ..., bold_rows: bool = ..., escape: bool = ...) -> None:
        ...
    
    def get_strcols(self) -> list[list[str]]:
        """
        Render a DataFrame to a list of columns (as lists of strings).
        """
        ...
    
    @property
    def should_show_dimensions(self) -> bool:
        ...
    
    @property
    def is_truncated(self) -> bool:
        ...
    
    @property
    def is_truncated_horizontally(self) -> bool:
        ...
    
    @property
    def is_truncated_vertically(self) -> bool:
        ...
    
    @property
    def dimensions_info(self) -> str:
        ...
    
    @property
    def has_index_names(self) -> bool:
        ...
    
    @property
    def has_column_names(self) -> bool:
        ...
    
    @property
    def show_row_idx_names(self) -> bool:
        ...
    
    @property
    def show_col_idx_names(self) -> bool:
        ...
    
    @property
    def max_rows_displayed(self) -> int:
        ...
    
    def truncate(self) -> None:
        """
        Check whether the frame should be truncated. If so, slice the frame up.
        """
        ...
    
    def format_col(self, i: int) -> list[str]:
        ...
    


class DataFrameRenderer:
    """Class for creating dataframe output in multiple formats.

    Called in pandas.core.generic.NDFrame:
        - to_csv
        - to_latex

    Called in pandas.core.frame.DataFrame:
        - to_html
        - to_string

    Parameters
    ----------
    fmt : DataFrameFormatter
        Formatter with the formatting options.
    """
    def __init__(self, fmt: DataFrameFormatter) -> None:
        ...
    
    def to_latex(self, buf: FilePath | WriteBuffer[str] | None = ..., column_format: str | None = ..., longtable: bool = ..., encoding: str | None = ..., multicolumn: bool = ..., multicolumn_format: str | None = ..., multirow: bool = ..., caption: str | tuple[str, str] | None = ..., label: str | None = ..., position: str | None = ...) -> str | None:
        """
        Render a DataFrame to a LaTeX tabular/longtable environment output.
        """
        ...
    
    def to_html(self, buf: FilePath | WriteBuffer[str] | None = ..., encoding: str | None = ..., classes: str | list | tuple | None = ..., notebook: bool = ..., border: int | bool | None = ..., table_id: str | None = ..., render_links: bool = ...) -> str | None:
        """
        Render a DataFrame to a html table.

        Parameters
        ----------
        buf : str, path object, file-like object, or None, default None
            String, path object (implementing ``os.PathLike[str]``), or file-like
            object implementing a string ``write()`` function. If None, the result is
            returned as a string.
        encoding : str, default “utf-8”
            Set character encoding.
        classes : str or list-like
            classes to include in the `class` attribute of the opening
            ``<table>`` tag, in addition to the default "dataframe".
        notebook : {True, False}, optional, default False
            Whether the generated HTML is for IPython Notebook.
        border : int
            A ``border=border`` attribute is included in the opening
            ``<table>`` tag. Default ``pd.options.display.html.border``.
        table_id : str, optional
            A css id is included in the opening `<table>` tag if specified.
        render_links : bool, default False
            Convert URLs to HTML links.
        """
        ...
    
    def to_string(self, buf: FilePath | WriteBuffer[str] | None = ..., encoding: str | None = ..., line_width: int | None = ...) -> str | None:
        """
        Render a DataFrame to a console-friendly tabular output.

        Parameters
        ----------
        buf : str, path object, file-like object, or None, default None
            String, path object (implementing ``os.PathLike[str]``), or file-like
            object implementing a string ``write()`` function. If None, the result is
            returned as a string.
        encoding: str, default “utf-8”
            Set character encoding.
        line_width : int, optional
            Width to wrap a line in characters.
        """
        ...
    
    @deprecate_kwarg(old_arg_name="line_terminator", new_arg_name="lineterminator")
    def to_csv(self, path_or_buf: FilePath | WriteBuffer[bytes] | WriteBuffer[str] | None = ..., encoding: str | None = ..., sep: str = ..., columns: Sequence[Hashable] | None = ..., index_label: IndexLabel | None = ..., mode: str = ..., compression: CompressionOptions = ..., quoting: int | None = ..., quotechar: str = ..., lineterminator: str | None = ..., chunksize: int | None = ..., date_format: str | None = ..., doublequote: bool = ..., escapechar: str | None = ..., errors: str = ..., storage_options: StorageOptions = ...) -> str | None:
        """
        Render dataframe as comma-separated file.
        """
        ...
    


def save_to_buffer(string: str, buf: FilePath | WriteBuffer[str] | None = ..., encoding: str | None = ...) -> str | None:
    """
    Perform serialization. Write to buf or return as string if buf is None.
    """
    ...

@contextmanager
def get_buffer(buf: FilePath | WriteBuffer[str] | None, encoding: str | None = ...) -> Iterator[WriteBuffer[str]] | Iterator[StringIO]:
    """
    Context manager to open, yield and close buffer for filenames or Path-like
    objects, otherwise yield buf unchanged.
    """
    ...

def format_array(values: Any, formatter: Callable | None, float_format: FloatFormatType | None = ..., na_rep: str = ..., digits: int | None = ..., space: str | int | None = ..., justify: str = ..., decimal: str = ..., leading_space: bool | None = ..., quoting: int | None = ...) -> list[str]:
    """
    Format an array for printing.

    Parameters
    ----------
    values
    formatter
    float_format
    na_rep
    digits
    space
    justify
    decimal
    leading_space : bool, optional, default True
        Whether the array should be formatted with a leading space.
        When an array as a column of a Series or DataFrame, we do want
        the leading space to pad between columns.

        When formatting an Index subclass
        (e.g. IntervalIndex._format_native_types), we don't want the
        leading space since it should be left-aligned.

    Returns
    -------
    List[str]
    """
    ...

class GenericArrayFormatter:
    def __init__(self, values: Any, digits: int = ..., formatter: Callable | None = ..., na_rep: str = ..., space: str | int = ..., float_format: FloatFormatType | None = ..., justify: str = ..., decimal: str = ..., quoting: int | None = ..., fixed_width: bool = ..., leading_space: bool | None = ...) -> None:
        ...
    
    def get_result(self) -> list[str]:
        ...
    


class FloatArrayFormatter(GenericArrayFormatter):
    def __init__(self, *args, **kwargs) -> None:
        ...
    
    def get_result_as_array(self) -> np.ndarray:
        """
        Returns the float values converted into strings using
        the parameters given at initialisation, as a numpy array
        """
        ...
    


class IntArrayFormatter(GenericArrayFormatter):
    ...


class Datetime64Formatter(GenericArrayFormatter):
    def __init__(self, values: np.ndarray | Series | DatetimeIndex | DatetimeArray, nat_rep: str = ..., date_format: None = ..., **kwargs) -> None:
        ...
    


class ExtensionArrayFormatter(GenericArrayFormatter):
    ...


def format_percentiles(percentiles: (np.ndarray | Sequence[float])) -> list[str]:
    """
    Outputs rounded and formatted percentiles.

    Parameters
    ----------
    percentiles : list-like, containing floats from interval [0,1]

    Returns
    -------
    formatted : list of strings

    Notes
    -----
    Rounding precision is chosen so that: (1) if any two elements of
    ``percentiles`` differ, they remain different after rounding
    (2) no entry is *rounded* to 0% or 100%.
    Any non-integer is always rounded to at least 1 decimal place.

    Examples
    --------
    Keeps all entries different after rounding:

    >>> format_percentiles([0.01999, 0.02001, 0.5, 0.666666, 0.9999])
    ['1.999%', '2.001%', '50%', '66.667%', '99.99%']

    No element is rounded to 0% or 100% (unless already equal to it).
    Duplicates are allowed:

    >>> format_percentiles([0, 0.5, 0.02001, 0.5, 0.666666, 0.9999])
    ['0%', '50%', '2.0%', '50%', '66.67%', '99.99%']
    """
    ...

def is_dates_only(values: np.ndarray | DatetimeArray | Index | DatetimeIndex) -> bool:
    ...

def get_format_datetime64(is_dates_only: bool, nat_rep: str = ..., date_format: str | None = ...) -> Callable:
    """Return a formatter callable taking a datetime64 as input and providing
    a string as output"""
    ...

def get_format_datetime64_from_values(values: np.ndarray | DatetimeArray | DatetimeIndex, date_format: str | None) -> str | None:
    """given values and a date_format, return a string format"""
    ...

class Datetime64TZFormatter(Datetime64Formatter):
    ...


class Timedelta64Formatter(GenericArrayFormatter):
    def __init__(self, values: np.ndarray | TimedeltaIndex, nat_rep: str = ..., box: bool = ..., **kwargs) -> None:
        ...
    


def get_format_timedelta64(values: np.ndarray | TimedeltaIndex | TimedeltaArray, nat_rep: str = ..., box: bool = ...) -> Callable:
    """
    Return a formatter function for a range of timedeltas.
    These will all have the same format argument

    If box, then show the return in quotes
    """
    ...

class EngFormatter:
    """
    Formats float values according to engineering format.

    Based on matplotlib.ticker.EngFormatter
    """
    ENG_PREFIXES = ...
    def __init__(self, accuracy: int | None = ..., use_eng_prefix: bool = ...) -> None:
        ...
    
    def __call__(self, num: float) -> str:
        """
        Formats a number in engineering notation, appending a letter
        representing the power of 1000 of the original number. Some examples:
        >>> format_eng = EngFormatter(accuracy=0, use_eng_prefix=True)
        >>> format_eng(0)
        ' 0'
        >>> format_eng = EngFormatter(accuracy=1, use_eng_prefix=True)
        >>> format_eng(1_000_000)
        ' 1.0M'
        >>> format_eng = EngFormatter(accuracy=2, use_eng_prefix=False)
        >>> format_eng("-1e-6")
        '-1.00E-06'

        @param num: the value to represent
        @type num: either a numeric value or a string that can be converted to
                   a numeric value (as per decimal.Decimal constructor)

        @return: engineering formatted string
        """
        ...
    


def set_eng_float_format(accuracy: int = ..., use_eng_prefix: bool = ...) -> None:
    """
    Alter default behavior on how float is formatted in DataFrame.
    Format float in engineering format. By accuracy, we mean the number of
    decimal digits after the floating point.

    See also EngFormatter.
    """
    ...

def get_level_lengths(levels: Any, sentinel: bool | object | str = ...) -> list[dict[int, int]]:
    """
    For each index in each level the function returns lengths of indexes.

    Parameters
    ----------
    levels : list of lists
        List of values on for level.
    sentinel : string, optional
        Value which states that no new index starts on there.

    Returns
    -------
    Returns list of maps. For each level returns map of indexes (key is index
    in row and value is length of index).
    """
    ...

def buffer_put_lines(buf: WriteBuffer[str], lines: list[str]) -> None:
    """
    Appends lines to a buffer.

    Parameters
    ----------
    buf
        The buffer to write to
    lines
        The lines to append.
    """
    ...




"""
This type stub file was generated by pyright.
"""

import numpy as np
from typing import Any
from pandas._typing import ExtensionDtype

def check_result_array(obj: object, dtype: np.dtype | ExtensionDtype) -> None:
    ...

def extract_result(res: object) -> Any:
    ...




"""
This type stub file was generated by pyright.
"""

def convert_to_line_delimits(s: str) -> str:
    """
    Helper function that converts JSON lists to line delimited JSON.
    """
    ...

def nested_to_record(ds, prefix: str = ..., sep: str = ..., level: int = ..., max_level: int | None = ...): # -> list[Unknown]:
    """
    A simplified json_normalize

    Converts a nested dict into a flat dict ("record"), unlike json_normalize,
    it does not attempt to extract a subset of the data.

    Parameters
    ----------
    ds : dict or list of dicts
    prefix: the prefix, optional, default: ""
    sep : str, default '.'
        Nested records will generate names separated by sep,
        e.g., for sep='.', { 'foo' : { 'bar' : 0 } } -> foo.bar
    level: int, optional, default: 0
        The number of levels in the json string.

    max_level: int, optional, default: None
        The max depth to normalize.

        .. versionadded:: 0.25.0

    Returns
    -------
    d - dict or list of dicts, matching `ds`

    Examples
    --------
    >>> nested_to_record(
    ...     dict(flat1=1, dict1=dict(c=1, d=2), nested=dict(e=dict(c=1, d=2), d=2))
    ... )
    {\
'flat1': 1, \
'dict1.c': 1, \
'dict1.d': 2, \
'nested.e.c': 1, \
'nested.e.d': 2, \
'nested.d': 2\
}
    """
    ...

json_normalize = ...



"""
This type stub file was generated by pyright.
"""

import numpy as np
from pandas.core.dtypes.base import register_extension_dtype
from pandas.core.arrays.numeric import NumericArray, NumericDtype

class IntegerDtype(NumericDtype):
    """
    An ExtensionDtype to hold a single size & kind of integer dtype.

    These specific implementations are subclasses of the non-public
    IntegerDtype. For example we have Int8Dtype to represent signed int 8s.

    The attributes name & type are set when these subclasses are created.
    """
    _default_np_dtype = ...
    _checker = ...
    @classmethod
    def construct_array_type(cls) -> type[IntegerArray]:
        """
        Return the array type associated with this dtype.

        Returns
        -------
        type
        """
        ...
    


class IntegerArray(NumericArray):
    """
    Array of integer (optional missing) values.

    .. versionchanged:: 1.0.0

       Now uses :attr:`pandas.NA` as the missing value rather
       than :attr:`numpy.nan`.

    .. warning::

       IntegerArray is currently experimental, and its API or internal
       implementation may change without warning.

    We represent an IntegerArray with 2 numpy arrays:

    - data: contains a numpy integer array of the appropriate dtype
    - mask: a boolean array holding a mask on the data, True is missing

    To construct an IntegerArray from generic array-like input, use
    :func:`pandas.array` with one of the integer dtypes (see examples).

    See :ref:`integer_na` for more.

    Parameters
    ----------
    values : numpy.ndarray
        A 1-d integer-dtype array.
    mask : numpy.ndarray
        A 1-d boolean-dtype array indicating missing values.
    copy : bool, default False
        Whether to copy the `values` and `mask`.

    Attributes
    ----------
    None

    Methods
    -------
    None

    Returns
    -------
    IntegerArray

    Examples
    --------
    Create an IntegerArray with :func:`pandas.array`.

    >>> int_array = pd.array([1, None, 3], dtype=pd.Int32Dtype())
    >>> int_array
    <IntegerArray>
    [1, <NA>, 3]
    Length: 3, dtype: Int32

    String aliases for the dtypes are also available. They are capitalized.

    >>> pd.array([1, None, 3], dtype='Int32')
    <IntegerArray>
    [1, <NA>, 3]
    Length: 3, dtype: Int32

    >>> pd.array([1, None, 3], dtype='UInt16')
    <IntegerArray>
    [1, <NA>, 3]
    Length: 3, dtype: UInt16
    """
    _dtype_cls = IntegerDtype
    _internal_fill_value = ...
    _truthy_value = ...
    _falsey_value = ...


_dtype_docstring = ...
@register_extension_dtype
class Int8Dtype(IntegerDtype):
    type = np.int8
    name = ...
    __doc__ = ...


@register_extension_dtype
class Int16Dtype(IntegerDtype):
    type = np.int16
    name = ...
    __doc__ = ...


@register_extension_dtype
class Int32Dtype(IntegerDtype):
    type = np.int32
    name = ...
    __doc__ = ...


@register_extension_dtype
class Int64Dtype(IntegerDtype):
    type = np.int64
    name = ...
    __doc__ = ...


@register_extension_dtype
class UInt8Dtype(IntegerDtype):
    type = np.uint8
    name = ...
    __doc__ = ...


@register_extension_dtype
class UInt16Dtype(IntegerDtype):
    type = np.uint16
    name = ...
    __doc__ = ...


@register_extension_dtype
class UInt32Dtype(IntegerDtype):
    type = np.uint32
    name = ...
    __doc__ = ...


@register_extension_dtype
class UInt64Dtype(IntegerDtype):
    type = np.uint64
    name = ...
    __doc__ = ...


INT_STR_TO_DTYPE: dict[str, IntegerDtype] = ...



"""
This type stub file was generated by pyright.
"""

from pandas._typing import DtypeObj

"""
Utility functions and objects for implementing the interchange API.
"""
class ArrowCTypes:
    """
    Enum for Apache Arrow C type format strings.

    The Arrow C data interface:
    https://arrow.apache.org/docs/format/CDataInterface.html#data-type-description-format-strings
    """
    NULL = ...
    BOOL = ...
    INT8 = ...
    UINT8 = ...
    INT16 = ...
    UINT16 = ...
    INT32 = ...
    UINT32 = ...
    INT64 = ...
    UINT64 = ...
    FLOAT16 = ...
    FLOAT32 = ...
    FLOAT64 = ...
    STRING = ...
    DATE32 = ...
    DATE64 = ...
    TIMESTAMP = ...
    TIME = ...


class Endianness:
    """Enum indicating the byte-order of a data-type."""
    LITTLE = ...
    BIG = ...
    NATIVE = ...
    NA = ...


def dtype_to_arrow_c_fmt(dtype: DtypeObj) -> str:
    """
    Represent pandas `dtype` as a format string in Apache Arrow C notation.

    Parameters
    ----------
    dtype : np.dtype
        Datatype of pandas DataFrame to represent.

    Returns
    -------
    str
        Format string in Apache Arrow C notation of the given `dtype`.
    """
    ...

class NoBufferPresent(Exception):
    """Exception to signal that there is no requested buffer."""
    ...





"""
This type stub file was generated by pyright.
"""

from abc import ABC, abstractmethod
from typing import Iterator, TYPE_CHECKING
from pandas.io.formats.format import DataFrameFormatter

"""
Module for formatting output data in Latex.
"""
if TYPE_CHECKING:
    ...
class RowStringConverter(ABC):
    r"""Converter for dataframe rows into LaTeX strings.

    Parameters
    ----------
    formatter : `DataFrameFormatter`
        Instance of `DataFrameFormatter`.
    multicolumn: bool, optional
        Whether to use \multicolumn macro.
    multicolumn_format: str, optional
        Multicolumn format.
    multirow: bool, optional
        Whether to use \multirow macro.

    """
    def __init__(self, formatter: DataFrameFormatter, multicolumn: bool = ..., multicolumn_format: str | None = ..., multirow: bool = ...) -> None:
        ...
    
    def get_strrow(self, row_num: int) -> str:
        """Get string representation of the row."""
        ...
    
    @property
    def index_levels(self) -> int:
        """Integer number of levels in index."""
        ...
    
    @property
    def column_levels(self) -> int:
        ...
    
    @property
    def header_levels(self) -> int:
        ...
    


class RowStringIterator(RowStringConverter):
    """Iterator over rows of the header or the body of the table."""
    @abstractmethod
    def __iter__(self) -> Iterator[str]:
        """Iterate over LaTeX string representations of rows."""
        ...
    


class RowHeaderIterator(RowStringIterator):
    """Iterator for the table header rows."""
    def __iter__(self) -> Iterator[str]:
        ...
    


class RowBodyIterator(RowStringIterator):
    """Iterator for the table body rows."""
    def __iter__(self) -> Iterator[str]:
        ...
    


class TableBuilderAbstract(ABC):
    """
    Abstract table builder producing string representation of LaTeX table.

    Parameters
    ----------
    formatter : `DataFrameFormatter`
        Instance of `DataFrameFormatter`.
    column_format: str, optional
        Column format, for example, 'rcl' for three columns.
    multicolumn: bool, optional
        Use multicolumn to enhance MultiIndex columns.
    multicolumn_format: str, optional
        The alignment for multicolumns, similar to column_format.
    multirow: bool, optional
        Use multirow to enhance MultiIndex rows.
    caption: str, optional
        Table caption.
    short_caption: str, optional
        Table short caption.
    label: str, optional
        LaTeX label.
    position: str, optional
        Float placement specifier, for example, 'htb'.
    """
    def __init__(self, formatter: DataFrameFormatter, column_format: str | None = ..., multicolumn: bool = ..., multicolumn_format: str | None = ..., multirow: bool = ..., caption: str | None = ..., short_caption: str | None = ..., label: str | None = ..., position: str | None = ...) -> None:
        ...
    
    def get_result(self) -> str:
        """String representation of LaTeX table."""
        ...
    
    @property
    @abstractmethod
    def env_begin(self) -> str:
        """Beginning of the environment."""
        ...
    
    @property
    @abstractmethod
    def top_separator(self) -> str:
        """Top level separator."""
        ...
    
    @property
    @abstractmethod
    def header(self) -> str:
        """Header lines."""
        ...
    
    @property
    @abstractmethod
    def middle_separator(self) -> str:
        """Middle level separator."""
        ...
    
    @property
    @abstractmethod
    def env_body(self) -> str:
        """Environment body."""
        ...
    
    @property
    @abstractmethod
    def bottom_separator(self) -> str:
        """Bottom level separator."""
        ...
    
    @property
    @abstractmethod
    def env_end(self) -> str:
        """End of the environment."""
        ...
    


class GenericTableBuilder(TableBuilderAbstract):
    """Table builder producing string representation of LaTeX table."""
    @property
    def header(self) -> str:
        ...
    
    @property
    def top_separator(self) -> str:
        ...
    
    @property
    def middle_separator(self) -> str:
        ...
    
    @property
    def env_body(self) -> str:
        ...
    


class LongTableBuilder(GenericTableBuilder):
    """Concrete table builder for longtable.

    >>> from pandas.io.formats import format as fmt
    >>> df = pd.DataFrame({"a": [1, 2], "b": ["b1", "b2"]})
    >>> formatter = fmt.DataFrameFormatter(df)
    >>> builder = LongTableBuilder(formatter, caption='a long table',
    ...                            label='tab:long', column_format='lrl')
    >>> table = builder.get_result()
    >>> print(table)
    \\begin{longtable}{lrl}
    \\caption{a long table}
    \\label{tab:long}\\\\
    \\toprule
    {} &  a &   b \\\\
    \\midrule
    \\endfirsthead
    \\caption[]{a long table} \\\\
    \\toprule
    {} &  a &   b \\\\
    \\midrule
    \\endhead
    \\midrule
    \\multicolumn{3}{r}{{Continued on next page}} \\\\
    \\midrule
    \\endfoot
    <BLANKLINE>
    \\bottomrule
    \\endlastfoot
    0 &  1 &  b1 \\\\
    1 &  2 &  b2 \\\\
    \\end{longtable}
    <BLANKLINE>
    """
    @property
    def env_begin(self) -> str:
        ...
    
    @property
    def middle_separator(self) -> str:
        ...
    
    @property
    def bottom_separator(self) -> str:
        ...
    
    @property
    def env_end(self) -> str:
        ...
    


class RegularTableBuilder(GenericTableBuilder):
    """Concrete table builder for regular table.

    >>> from pandas.io.formats import format as fmt
    >>> df = pd.DataFrame({"a": [1, 2], "b": ["b1", "b2"]})
    >>> formatter = fmt.DataFrameFormatter(df)
    >>> builder = RegularTableBuilder(formatter, caption='caption', label='lab',
    ...                               column_format='lrc')
    >>> table = builder.get_result()
    >>> print(table)
    \\begin{table}
    \\centering
    \\caption{caption}
    \\label{lab}
    \\begin{tabular}{lrc}
    \\toprule
    {} &  a &   b \\\\
    \\midrule
    0 &  1 &  b1 \\\\
    1 &  2 &  b2 \\\\
    \\bottomrule
    \\end{tabular}
    \\end{table}
    <BLANKLINE>
    """
    @property
    def env_begin(self) -> str:
        ...
    
    @property
    def bottom_separator(self) -> str:
        ...
    
    @property
    def env_end(self) -> str:
        ...
    


class TabularBuilder(GenericTableBuilder):
    """Concrete table builder for tabular environment.

    >>> from pandas.io.formats import format as fmt
    >>> df = pd.DataFrame({"a": [1, 2], "b": ["b1", "b2"]})
    >>> formatter = fmt.DataFrameFormatter(df)
    >>> builder = TabularBuilder(formatter, column_format='lrc')
    >>> table = builder.get_result()
    >>> print(table)
    \\begin{tabular}{lrc}
    \\toprule
    {} &  a &   b \\\\
    \\midrule
    0 &  1 &  b1 \\\\
    1 &  2 &  b2 \\\\
    \\bottomrule
    \\end{tabular}
    <BLANKLINE>
    """
    @property
    def env_begin(self) -> str:
        ...
    
    @property
    def bottom_separator(self) -> str:
        ...
    
    @property
    def env_end(self) -> str:
        ...
    


class LatexFormatter:
    r"""
    Used to render a DataFrame to a LaTeX tabular/longtable environment output.

    Parameters
    ----------
    formatter : `DataFrameFormatter`
    longtable : bool, default False
        Use longtable environment.
    column_format : str, default None
        The columns format as specified in `LaTeX table format
        <https://en.wikibooks.org/wiki/LaTeX/Tables>`__ e.g 'rcl' for 3 columns
    multicolumn : bool, default False
        Use \multicolumn to enhance MultiIndex columns.
    multicolumn_format : str, default 'l'
        The alignment for multicolumns, similar to `column_format`
    multirow : bool, default False
        Use \multirow to enhance MultiIndex rows.
    caption : str or tuple, optional
        Tuple (full_caption, short_caption),
        which results in \caption[short_caption]{full_caption};
        if a single string is passed, no short caption will be set.
    label : str, optional
        The LaTeX label to be placed inside ``\label{}`` in the output.
    position : str, optional
        The LaTeX positional argument for tables, to be placed after
        ``\begin{}`` in the output.

    See Also
    --------
    HTMLFormatter
    """
    def __init__(self, formatter: DataFrameFormatter, longtable: bool = ..., column_format: str | None = ..., multicolumn: bool = ..., multicolumn_format: str | None = ..., multirow: bool = ..., caption: str | tuple[str, str] | None = ..., label: str | None = ..., position: str | None = ...) -> None:
        ...
    
    def to_string(self) -> str:
        """
        Render a DataFrame to a LaTeX tabular, longtable, or table/tabular
        environment output.
        """
        ...
    
    @property
    def builder(self) -> TableBuilderAbstract:
        """Concrete table builder.

        Returns
        -------
        TableBuilder
        """
        ...
    
    @property
    def column_format(self) -> str | None:
        """Column format."""
        ...
    
    @column_format.setter
    def column_format(self, input_column_format: str | None) -> None:
        """Setter for column format."""
        ...
    


if __name__ == "__main__":
    ...



"""
This type stub file was generated by pyright.
"""

import pandas as pd
from pandas.util._decorators import cache_readonly
from pandas.core.interchange.dataframe_protocol import Column, ColumnBuffers, DtypeKind

_NP_KINDS = ...
_NULL_DESCRIPTION = ...
_NO_VALIDITY_BUFFER = ...
class PandasColumn(Column):
    """
    A column object, with only the methods and properties required by the
    interchange protocol defined.
    A column can contain one or more chunks. Each chunk can contain up to three
    buffers - a data buffer, a mask buffer (depending on null representation),
    and an offsets buffer (if variable-size binary; e.g., variable-length
    strings).
    Note: this Column object can only be produced by ``__dataframe__``, so
          doesn't need its own version or ``__column__`` protocol.
    """
    def __init__(self, column: pd.Series, allow_copy: bool = ...) -> None:
        """
        Note: doesn't deal with extension arrays yet, just assume a regular
        Series/ndarray for now.
        """
        ...
    
    def size(self) -> int:
        """
        Size of the column, in elements.
        """
        ...
    
    @property
    def offset(self) -> int:
        """
        Offset of first element. Always zero.
        """
        ...
    
    @cache_readonly
    def dtype(self) -> tuple[DtypeKind, int, str, str]:
        ...
    
    @property
    def describe_categorical(self): # -> dict[str, Unknown]:
        """
        If the dtype is categorical, there are two options:
        - There are only values in the data buffer.
        - There is a separate non-categorical Column encoding for categorical values.

        Raises TypeError if the dtype is not categorical

        Content of returned dict:
            - "is_ordered" : bool, whether the ordering of dictionary indices is
                             semantically meaningful.
            - "is_dictionary" : bool, whether a dictionary-style mapping of
                                categorical values to other objects exists
            - "categories" : Column representing the (implicit) mapping of indices to
                             category values (e.g. an array of cat1, cat2, ...).
                             None if not a dictionary-style categorical.
        """
        ...
    
    @property
    def describe_null(self): # -> tuple[Unknown, Unknown]:
        ...
    
    @cache_readonly
    def null_count(self) -> int:
        """
        Number of null elements. Should always be known.
        """
        ...
    
    @property
    def metadata(self) -> dict[str, pd.Index]:
        """
        Store specific metadata of the column.
        """
        ...
    
    def num_chunks(self) -> int:
        """
        Return the number of chunks the column consists of.
        """
        ...
    
    def get_chunks(self, n_chunks: int | None = ...): # -> Generator[PandasColumn | Self@PandasColumn, None, None]:
        """
        Return an iterator yielding the chunks.
        See `DataFrame.get_chunks` for details on ``n_chunks``.
        """
        ...
    
    def get_buffers(self) -> ColumnBuffers:
        """
        Return a dictionary containing the underlying buffers.
        The returned dictionary has the following contents:
            - "data": a two-element tuple whose first element is a buffer
                      containing the data and whose second element is the data
                      buffer's associated dtype.
            - "validity": a two-element tuple whose first element is a buffer
                          containing mask values indicating missing data and
                          whose second element is the mask value buffer's
                          associated dtype. None if the null representation is
                          not a bit or byte mask.
            - "offsets": a two-element tuple whose first element is a buffer
                         containing the offset values for variable-size binary
                         data (e.g., variable-length strings) and whose second
                         element is the offsets buffer's associated dtype. None
                         if the data buffer does not have an associated offsets
                         buffer.
        """
        ...
    





# BoARIO : The Adaptative Regional Input Output model in python.
# Copyright (C) 2022  Samuel Juhel
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

'''
Simulation module

This module defines the Simulation object, which represent a BoARIO simulation environment.

'''

from __future__ import annotations
import json
import pickle
import pathlib
from typing import Optional, Union
import logging
import math
import numpy as np
import progressbar
import pymrio as pym
from pymrio.core.mriosystem import IOSystem

from boario.event import Event
from boario.model_base import ARIOBaseModel
from boario.extended_models import ARIOModelPsi
from boario import logger
from boario import DEBUGFORMATTER

__all__=['Simulation']

class Simulation(object):
    """Defines a simulation object with a set of parameters and an IOSystem.

    This class wraps an :class:`~boario.model_base.ARIOBaseModel` or :class:`~boario.extended_models.ARIOModelPsi`, and create the context for
    simulations using this model. It stores execution parameters as well as events perturbing
    the model.

    Attributes
    ----------
    params : dict
        Parameters to run the simulation with. If str or Path, it must lead
        to a json file containing a dictionary of the parameters.

    results_storage : pathlib.Path
        Path to store the results to.

    model : Union[ARIOBaseModel, ARIOModelPsi]
        The model to run the simulation with.

    current_temporal_unit : int
        Tracks the number of `temporal_units` elapsed since simulation start.
        This may differs from the number of `steps` if the parameter `temporal_units_by_step` differs from 1 temporal_unit as `current_temporal_unit` is actually `step` * `temporal_units_by_step`.

    n_temporal_units_to_sim : int
        The total number of `temporal_units` to simulate.

    events : list[Event]
        The list of events to shock the model with during the simulation.

    current_events : list[Event]
        The list of events that

    Raises
    ------
    TypeError
        This error is raised when parameters files (for either the simulation or the mrio table) is not of a correct type.

    FileNotFoundError
        This error is raised when one of the required file to initialize
        the simulation was not found and should print out which one.
    """
    def __init__(self, params: Union[dict, str, pathlib.Path], mrio_system: Union[IOSystem, str, pathlib.Path, None] = None, mrio_params: Optional[dict] = None, modeltype:str = "ARIOPsi") -> None:
        """Initialisation of a Simulation object uses these parameters

        Parameters
        ----------
        params : Union[dict, str, pathlib.Path]
            Parameters to run the simulation with. If str or Path, it must lead
            to a json file containing a dictionary of the parameters.
        mrio_system : Union[IOSystem, str, pathlib.Path, None], default: None
            ``pymrio.IOSystem`` to run the simulation with. If str or Path, it must
            lead to either :

            1) a pickle file of an IOSystem that was previously
            generated with the pymrio package (be careful that such files are
            not cross-system compatible) or,
            2) a directory loadable with ``pymrio.load_all`` (see `Loading, saving and exporting pymrio data <https://pymrio.readthedocs.io/en/latest/notebooks/load_save_export.html>`_).
        """
        logger.info("Initializing new simulation instance")
        super().__init__()

        # SIMULATION PARAMETER LOADING
        simulation_params = None
        params_path = None

        # CASE DICT
        if isinstance(params, dict):
            logger.info("Loading simulation parameters from dict")
            simulation_params = params

        # CASE STR|PATH
        elif isinstance(params, str):
            params_path=pathlib.Path(params)
        elif isinstance(params, pathlib.Path):
            params_path=params

        # OTHERWISE ERROR
        else:
            raise TypeError("params must be either a dict, a str or a pathlib.Path, not a {}".format(type(params)))

        # IF NOT DEFINED, then its a path to the file, or a directory with the file.
        if simulation_params is None:
            if params_path is None:
                raise ValueError("Parameters path is set yet, this is a problem.")
            if params_path.is_dir():
                simulation_params_path = params_path / 'params.json'
                if not simulation_params_path.exists():
                    raise FileNotFoundError("Simulation parameters file not found, it should be here: ",simulation_params_path.absolute())
                else:
                    with simulation_params_path.open() as f:
                        logger.info("Loading simulation parameters from {}".format(simulation_params_path))
                        simulation_params = json.load(f)
            else:
                if not params_path.exists():
                    raise FileNotFoundError("Simulation parameters file not found, it should be here: ",params_path.absolute())
                else:
                    with params_path.open() as f:
                        logger.info("Loading simulation parameters from {}".format(params_path))
                        simulation_params = json.load(f)

        # MRIO PARAMETERS LOADING
        mrio_params_loaded = None
        mrio_params_path = None
        # CASE: given as parameter
        if mrio_params is not None:
            # Is it directly a dict ?
            if isinstance(mrio_params, dict):
                logger.info("Loading MRIO parameters from dict")
                mrio_params_loaded = mrio_params
            # If it is a str, convert to path
            elif isinstance(mrio_params, str):
                mrio_params_path=pathlib.Path(mrio_params)
            # Already a path
            elif isinstance(mrio_params, pathlib.Path):
                mrio_params_path=mrio_params
            # Else error !
            else:
                raise TypeError("params must be either a dict, a str or a pathlib.Path, not a {}".format(type(params)))

            # Still in the case where it is an argument
            if mrio_params_loaded is None:
                if params_path is not None and params_path.is_dir():
                    mrio_params_path = params_path / 'mrio_params.json'
                    if not mrio_params_path.exists():
                        raise FileNotFoundError("MRIO parameters file not found, it should be here: ",mrio_params_path.absolute())
                    else:
                        with mrio_params_path.open() as f:
                            logger.info("Loading MRIO parameters from {}".format(mrio_params_path))
                            mrio_params_loaded = json.load(f)
                else:
                    if mrio_params_path is None:
                        raise ValueError("MRIO parameters are not set a this point and this is a problem.")
                    if not mrio_params_path.exists():
                        raise FileNotFoundError("MRIO parameters file not found, it should be here: ",mrio_params_path.absolute())
                    else:
                        with mrio_params_path.open() as f:
                            logger.info("Loading MRIO parameters from {}".format(mrio_params_path))
                            mrio_params_loaded = json.load(f)

        # In case it is not given as an argument :
        elif 'mrio_params_file' not in simulation_params.keys():
            raise FileNotFoundError("Unable to load MRIO parameters file from arguments and the path to it is not set in the simulation parameters file")
        else:
            mrio_params_path = pathlib.Path(simulation_params['mrio_params_file'])
            if not mrio_params_path.exists():
                logger.warning('MRIO parameters file specified in simulation params was not found: {}'.format(mrio_params_path))
            else:
                with mrio_params_path.open() as f:
                    mrio_params_loaded = json.load(f)

        # Probably useless now
        if mrio_params_loaded is None:
            raise FileNotFoundError("Couldn't load the MRIO params (tried with simulation params and the default file)")

        # LOAD MRIO system
        if isinstance(mrio_system, str):
            mrio_system = pathlib.Path(mrio_system)
        if isinstance(mrio_system, pathlib.Path):
            if not mrio_system.exists():
                raise FileNotFoundError("This file does not exist: ",mrio_system)
            if mrio_system.suffix == '.pkl':
                with mrio_system.open(mode='rb') as fp:
                    mrio = pickle.load(fp)
            else:
                if not mrio_system.is_dir():
                    raise FileNotFoundError("This file should be a pickle file or a directory loadable by pymrio: ", mrio_system.absolute())
                mrio = pym.load_all(mrio_system.absolute())
        elif isinstance(mrio_system, IOSystem):
            mrio = mrio_system
        else:
            raise TypeError("mrio_system must be either a str, a pathlib.Path or an pymrio.IOSystem, not a %s", str(type(mrio_system)))

        if not isinstance(mrio, IOSystem):
            raise TypeError("At this point, mrio should be an IOSystem, not a %s", str(type(mrio)))

        self.params = simulation_params
        self.results_storage = results_storage = pathlib.Path(self.params['output_dir']+"/"+self.params['results_storage'])
        if not results_storage.exists():
            results_storage.mkdir(parents=True)
        if modeltype == "ARIOBase":
            logger.warning("Be aware you are using an ARIOBaseModel instance, which is mostly intended for testing and developping and not using for economic modeling (it has numerous numeric instabilities among other things). You should use ARIOModelPsi for this purpose.")
            self.model = ARIOBaseModel(mrio, mrio_params_loaded, simulation_params, results_storage)
        elif modeltype == "ARIOPsi":
            self.model = ARIOModelPsi(mrio, mrio_params_loaded, simulation_params, results_storage)
        else:
            raise NotImplementedError("""This model type is not implemented : {}
Available types are {}
            """.format(modeltype,str(["ARIOBase","ARIOPsi"])))
        self.all_events = []
        self.currently_happening_events = []
        self.events_timings = set()
        self.n_temporal_units_to_sim = simulation_params['n_temporal_units_to_sim']
        self.current_temporal_unit = 0
        self.equi = {
            (int(0),int(0),"production"):"equi",
            (int(0),int(0),"stocks"):"equi",
            (int(0),int(0),"rebuilding"):"equi"
        }
        self.n_temporal_units_simulated = 0
        self._n_checks = 0
        self._monotony_checker = 0
        self.scheme = 'proportional'
        self.has_crashed = False
        logger.info("Initialized !")

    def loop(self, progress:bool=True):
        r"""Launch the simulation loop.

        This method launch the simulation for the number of steps to simulate
        described by the attribute ``n_temporal_units_to_sim``, calling the
        :meth:`next_step` method. For convenience, it dumps the
        parameters used in the logs just before running the loop. Once the loop
        is completed, it flushes the different memmaps generated.

        Parameters
        ----------

        progress: bool, default: True
            If True show a progress bar of the loop in the console.
        """
        logger.info("Starting model loop for at most {} steps".format(self.n_temporal_units_to_sim//self.model.n_temporal_units_by_step+1))
        logger.info("One step is {}/{} of a year".format(self.model.n_temporal_units_by_step, self.model.iotable_year_to_temporal_unit_factor))
        tmp = logging.FileHandler(self.results_storage/"simulation.log")
        tmp.setLevel(logging.DEBUG)
        tmp.setFormatter(DEBUGFORMATTER)
        logger.addHandler(tmp)
        logger.info("Events : {}".format(self.all_events))
        with (pathlib.Path(self.params["output_dir"]+"/"+self.params['results_storage'])/"simulated_events.json").open('w') as f:
            event_dicts = [ev.event_dict for ev in self.all_events]
            json.dump(event_dicts, f, indent=4)
        if progress:
            widgets = [
                'Processed: ', progressbar.Counter('Step: %(value)d'), ' ~ ', progressbar.Percentage(), ' ', progressbar.ETA(),
            ]
            bar = progressbar.ProgressBar(widgets=widgets, redirect_stdout=True)
            for _ in bar(range(0,self.n_temporal_units_to_sim,math.floor(self.params['temporal_units_by_step']))):
                #assert self.current_temporal_unit == t
                step_res = self.next_step()
                self.n_temporal_units_simulated = self.current_temporal_unit
                if step_res == 1:
                    self.has_crashed = True
                    logger.warning(f"""Economy seems to have crashed.
                    - At step : {self.current_temporal_unit}
                    """
                    )
                    break
                elif self._monotony_checker > 3:
                    logger.warning(f"""Economy seems to have found an equilibrium
                    - At step : {self.current_temporal_unit}
                    """
                    )
                    break
        else:
            for _ in range(0,self.n_temporal_units_to_sim,math.floor(self.params['temporal_units_by_step'])):
                #assert self.current_temporal_unit == t
                step_res = self.next_step()
                self.n_temporal_units_simulated = self.current_temporal_unit
                if step_res == 1:
                    self.has_crashed = True
                    logger.warning(f"""Economy seems to have crashed.
                    - At step : {self.current_temporal_unit}
                    """
                    )
                    break
                elif self._monotony_checker > 3:
                    logger.warning(f"""Economy seems to have found an equilibrium
                    - At step : {self.current_temporal_unit}
                    """
                    )
                    break

        self.model.rebuild_demand_evolution.flush()
        self.model.final_demand_unmet_evolution.flush()
        self.model.final_demand_evolution.flush()
        self.model.io_demand_evolution.flush()
        self.model.production_evolution.flush()
        self.model.limiting_stocks_evolution.flush()
        self.model.rebuild_production_evolution.flush()
        if self.params['register_stocks']:
            self.model.stocks_evolution.flush()
        self.model.overproduction_evolution.flush()
        self.model.production_cap_evolution.flush()
        self.params['n_temporal_units_simulated'] = self.n_temporal_units_simulated
        self.params['has_crashed'] = self.has_crashed
        with (pathlib.Path(self.params["output_dir"]+"/"+self.params['results_storage'])/"simulated_params.json").open('w') as f:
            json.dump(self.params, f, indent=4)
        with (pathlib.Path(self.params["output_dir"]+"/"+self.params['results_storage'])/"equilibrium_checks.json").open('w') as f:
            json.dump({str(k): v for k, v in self.equi.items()}, f, indent=4)
        logger.info('Loop complete')
        if progress:
            bar.finish() # type: ignore (bar possibly unbound but actually not possible)

    def next_step(self, check_period : int = 182, min_steps_check : Optional[int] = None, min_failing_regions : Optional[int] = None):
        """Advance the model run by one step.

        This method wraps all computations and logging to proceed to the next
        step of the simulation run. First it checks if an event is planned to
        occur at the current step and if so, shocks the model with the
        corresponding event. Then it :

        1) Computes the production required by demand (using :meth:`~boario.model_base.ARIOBaseModel.calc_prod_reqby_demand`)

        2) Computes the production capacity vector of the current step (using :meth:`~boario.model_base.ARIOBaseModel.calc_production_cap`)

        3) Computes the actual production vector for the step (using :meth:`~boario.model_base.ARIOBaseModel.calc_production`)

        4) Distribute the actual production towards the different demands (intermediate, final, rebuilding) and the changes in the stocks matrix (using :meth:`~boario.model_base.ARIOBaseModel.distribute_production`)

        5) Computes the orders matrix for the next step (using :meth:`~boario.model_base.ARIOBaseModel.calc_orders`)

        6) Computes the new overproduction vector for the next step (using :meth:`~boario.model_base.ARIOBaseModel.calc_overproduction`)

        See :ref:`Mathematical background <boario-math>` section for more in depth information.

        Parameters
        ----------

        check_period : int, default: 10
            [Deprecated] Number of steps between each crash/equilibrium checking.

        min_steps_check : int, default: None
            [Deprecated] Minimum number of steps before checking for crash/equilibrium. If none, it is set to a fifth of the number of steps to simulate.

        min_failing_regions : int, default: None
            [Deprecated] Minimum number of 'failing regions' required to consider the economy has 'crashed' (see :func:`~ario3.mriosystem.MrioSystem.check_crash`:).

        """
        if min_steps_check is None:
            min_steps_check = self.n_temporal_units_to_sim // 5
        if min_failing_regions is None:
            min_failing_regions = self.model.n_regions*self.model.n_sectors // 3

        # Check if there are new events to add,
        # if some happening events can start rebuilding (if rebuildable),
        # and updates the internal model production_cap decrease and rebuild_demand
        self.check_happening_events()

        if self.params['register_stocks']:
            self.model.write_stocks(self.current_temporal_unit)
        if self.current_temporal_unit > 1:
                self.model.calc_overproduction()
        self.model.write_overproduction(self.current_temporal_unit)
        self.model.write_rebuild_demand(self.current_temporal_unit)
        self.model.write_final_demand(self.current_temporal_unit)
        self.model.write_io_demand(self.current_temporal_unit)
        constraints = self.model.calc_production(self.current_temporal_unit)
        self.model.write_limiting_stocks(self.current_temporal_unit, constraints)
        self.model.write_production(self.current_temporal_unit)
        self.model.write_production_max(self.current_temporal_unit)
        try:
            rebuildable_events = [ev for ev in self.currently_happening_events if ev.rebuildable]
            events_to_remove = self.model.distribute_production(self.current_temporal_unit, rebuildable_events, self.scheme)
        except RuntimeError as e:
            logger.exception("This exception happened:",e)
            return 1
        if events_to_remove != []:
            self.currently_happening_events = [e for e in self.currently_happening_events if e not in events_to_remove]
            for e in events_to_remove:
                logger.info("Temporal_Unit : {} ~ Event named {} that occured at {} in {} for {} damages is completely rebuilt".format(self.current_temporal_unit, e.name,e.occurrence, e.aff_regions, e.total_kapital_destroyed))

        self.model.calc_orders()
        # TODO : Redo this properly
        n_checks = self.current_temporal_unit // check_period
        if (n_checks > self._n_checks):
            self.check_equilibrium(n_checks)
            self._n_checks+=1

        self.current_temporal_unit += self.params['temporal_units_by_step']
        return 0

    def check_equilibrium(self,n_checks:int):
        if np.greater_equal(self.model.production,self.model.X_0).all():
            self.equi[(n_checks,self.current_temporal_unit,"production")] = "greater"
        elif np.allclose(self.model.production,self.model.X_0,atol=0.01):
            self.equi[(n_checks,self.current_temporal_unit,"production")] = "equi"
        else:
            self.equi[(n_checks,self.current_temporal_unit,"production")] = "not equi"

        if np.greater_equal(self.model.matrix_stock,self.model.matrix_stock_0).all():
            self.equi[(n_checks,self.current_temporal_unit,"stocks")] = "greater"
        elif np.allclose(self.model.production,self.model.X_0,atol=0.01):
            self.equi[(n_checks,self.current_temporal_unit,"stocks")] = "equi"
        else:
            self.equi[(n_checks,self.current_temporal_unit,"stocks")] = "not equi"

        if self.model.tot_rebuild_demand is None or not self.model.tot_rebuild_demand.any():
            self.equi[(n_checks,self.current_temporal_unit,"rebuilding")] = "finished"
        else:
            self.equi[(n_checks,self.current_temporal_unit,"rebuilding")] = "not finished"

    def read_events_from_list(self, events_list : list[dict]):
        """Import a list of events (as a list of dictionaries) into the model.

        Also performs various checks on the events to avoid badly written events.
        See :ref:`How to define Events <boario-events>` to understand how to write events dictionaries or JSON files.

        Parameters
        ----------
        events_list :
            List of events as dictionaries.

        """
        logger.info("Reading events from given list and adding them to the model")
        if not isinstance(events_list, list):
            if isinstance(events_list, dict):
                raise TypeError("read_events_from_list() takes a list of event dicts as an argument, not a single event dict, you might want to use read_event(your_event) instead.")
            else:
                raise TypeError("read_events_from_list() takes a list of event dicts as an argument, not a {}".format(type(events_list)))
        for ev_dic in events_list:
            self.read_event(ev_dic)

    def read_event(self, ev_dic:dict):
        if ev_dic['aff_sectors'] == 'all':
            ev_dic['aff_sectors'] = list(self.model.sectors)
        ev = Event(ev_dic)
        self.all_events.append(ev)
        self.events_timings.add(ev.occurrence)


    def reset_sim_with_same_events(self):
        """Resets the model to its initial status (without removing the events).
        """

        logger.info('Resetting model to initial status (with same events)')
        self.current_temporal_unit = 0
        self._monotony_checker = 0
        self.n_temporal_units_simulated = 0
        self.has_crashed = False
        self.model.reset_module(self.params)

    def reset_sim_full(self):
        """Resets the model to its initial status and remove all events.
        """

        self.reset_sim_with_same_events()
        logger.info('Resetting events')
        self.all_events = []
        self.events_timings = set()

    def update_params(self, new_params:dict):
        """Update the parameters of the model.

        Replace the ``params`` attribute with ``new_params`` and logs the update.
        This method also checks if the directory specified to save the results exists and create it otherwise.

        .. warning::
            Be aware this method calls :meth:`~boario.model_base.ARIOBaseModel.update_params`, which resets the memmap files located in the results directory !

        Parameters
        ----------
        new_params : dict
            New dictionnary of parameters to use.

        """
        logger.info('Updating model parameters')
        self.params = new_params
        results_storage = pathlib.Path(self.params['output_dir']+"/"+self.params['results_storage'])
        if not results_storage.exists():
            results_storage.mkdir(parents=True)
        self.model.update_params(self.params)

    def write_index(self, index_file:Union[str, pathlib.Path]):
        """Write the index of the dataframes used in the model in a json file.

        See :meth:`~boario.model_base.ARIOBaseModel.write_index` for a more detailed documentation.

        Parameters
        ----------
        index_file : Union[str, pathlib.Path]
            name of the file to save the indexes to.

        """

        self.model.write_index(index_file)

    def read_events(self, events_file:Union[str, pathlib.Path]):
        """Read events from a json file.

        Parameters
        ----------
        events_file :
            path to a json file

        Raises
        ------
        FileNotFoundError
            If file does not exist

        """
        logger.info("Reading events from {} and adding them to the model".format(events_file))
        if isinstance(events_file,str):
            events_file = pathlib.Path(events_file)
        elif not isinstance(events_file,pathlib.Path):
            raise TypeError("Given index file is not an str or a Path")
        if not events_file.exists():
            raise FileNotFoundError("This file does not exist: ",events_file)
        else:
            with events_file.open('r') as f:
                events = json.load(f)
        if isinstance(events,list):
            self.read_events_from_list(events)
        else:
            self.read_events_from_list([events])

    def check_happening_events(self) -> None:
        for ev in self.all_events:
            if not ev.happened:
                if ((self.current_temporal_unit-self.params['temporal_units_by_step']) <= ev.occurrence <= self.current_temporal_unit):
                    logger.info("Temporal_Unit : {} ~ Shocking model with new event".format(self.current_temporal_unit))
                    logger.info("Affected regions are : {}".format(ev.aff_regions))
                    ev.happened = True
                    self.currently_happening_events.append(ev)
        for ev in self.currently_happening_events:
            ev.rebuildable = self.current_temporal_unit
            ev.recoverable = self.current_temporal_unit
            if ev.recoverable:
                ev.recovery(self.current_temporal_unit)
        self.model.update_system_from_events(self.currently_happening_events)


#    def check_current_events(self)-> None:
#        if self.current_events != []:
#            for ev in self.current_events:
#                pass



"""
This type stub file was generated by pyright.
"""

import numpy as np
from datetime import timedelta
from typing import Literal, Sequence, TYPE_CHECKING, TypeVar, overload
from pandas._libs.tslibs import BaseOffset, period as libperiod
from pandas._libs.tslibs.offsets import Tick
from pandas._libs.tslibs.period import Period
from pandas._typing import AnyArrayLike, Dtype, NpDtype, NumpySorter, NumpyValueArrayLike, npt
from pandas.util._decorators import cache_readonly, doc
from pandas.core.dtypes.dtypes import PeriodDtype
from pandas.core.arrays import DatetimeArray, datetimelike as dtl
from pandas.core.arrays.base import ExtensionArray

if TYPE_CHECKING:
    ...
BaseOffsetT = TypeVar("BaseOffsetT", bound=BaseOffset)
_shared_doc_kwargs = ...
class PeriodArray(dtl.DatelikeOps, libperiod.PeriodMixin):
    """
    Pandas ExtensionArray for storing Period data.

    Users should use :func:`~pandas.period_array` to create new instances.
    Alternatively, :func:`~pandas.array` can be used to create new instances
    from a sequence of Period scalars.

    Parameters
    ----------
    values : Union[PeriodArray, Series[period], ndarray[int], PeriodIndex]
        The data to store. These should be arrays that can be directly
        converted to ordinals without inference or copy (PeriodArray,
        ndarray[int64]), or a box around such an array (Series[period],
        PeriodIndex).
    dtype : PeriodDtype, optional
        A PeriodDtype instance from which to extract a `freq`. If both
        `freq` and `dtype` are specified, then the frequencies must match.
    freq : str or DateOffset
        The `freq` to use for the array. Mostly applicable when `values`
        is an ndarray of integers, when `freq` is required. When `values`
        is a PeriodArray (or box around), it's checked that ``values.freq``
        matches `freq`.
    copy : bool, default False
        Whether to copy the ordinals before storing.

    Attributes
    ----------
    None

    Methods
    -------
    None

    See Also
    --------
    Period: Represents a period of time.
    PeriodIndex : Immutable Index for period data.
    period_range: Create a fixed-frequency PeriodArray.
    array: Construct a pandas array.

    Notes
    -----
    There are two components to a PeriodArray

    - ordinals : integer ndarray
    - freq : pd.tseries.offsets.Offset

    The values are physically stored as a 1-D ndarray of integers. These are
    called "ordinals" and represent some kind of offset from a base.

    The `freq` indicates the span covered by each element of the array.
    All elements in the PeriodArray have the same `freq`.
    """
    __array_priority__ = ...
    _typ = ...
    _internal_fill_value = ...
    _recognized_scalars = ...
    _is_recognized_dtype = ...
    _infer_matches = ...
    _other_ops: list[str] = ...
    _bool_ops: list[str] = ...
    _object_ops: list[str] = ...
    _field_ops: list[str] = ...
    _datetimelike_ops: list[str] = ...
    _datetimelike_methods: list[str] = ...
    _dtype: PeriodDtype
    def __init__(self, values, dtype: Dtype | None = ..., freq=..., copy: bool = ...) -> None:
        ...
    
    @cache_readonly
    def dtype(self) -> PeriodDtype:
        ...
    
    @property
    def freq(self) -> BaseOffset:
        """
        Return the frequency object for this PeriodArray.
        """
        ...
    
    def __array__(self, dtype: NpDtype | None = ...) -> np.ndarray:
        ...
    
    def __arrow_array__(self, type=...):
        """
        Convert myself into a pyarrow Array.
        """
        ...
    
    year = ...
    month = ...
    day = ...
    hour = ...
    minute = ...
    second = ...
    weekofyear = ...
    week = ...
    day_of_week = ...
    dayofweek = ...
    weekday = ...
    dayofyear = ...
    quarter = ...
    qyear = ...
    days_in_month = ...
    daysinmonth = ...
    @property
    def is_leap_year(self) -> np.ndarray:
        """
        Logical indicating if the date belongs to a leap year.
        """
        ...
    
    def to_timestamp(self, freq=..., how: str = ...) -> DatetimeArray:
        """
        Cast to DatetimeArray/Index.

        Parameters
        ----------
        freq : str or DateOffset, optional
            Target frequency. The default is 'D' for week or longer,
            'S' otherwise.
        how : {'s', 'e', 'start', 'end'}
            Whether to use the start or end of the time period being converted.

        Returns
        -------
        DatetimeArray/Index
        """
        ...
    
    @doc(**_shared_doc_kwargs, other="PeriodIndex", other_name="PeriodIndex")
    def asfreq(self, freq=..., how: str = ...) -> PeriodArray:
        """
        Convert the {klass} to the specified frequency `freq`.

        Equivalent to applying :meth:`pandas.Period.asfreq` with the given arguments
        to each :class:`~pandas.Period` in this {klass}.

        Parameters
        ----------
        freq : str
            A frequency.
        how : str {{'E', 'S'}}, default 'E'
            Whether the elements should be aligned to the end
            or start within pa period.

            * 'E', 'END', or 'FINISH' for end,
            * 'S', 'START', or 'BEGIN' for start.

            January 31st ('END') vs. January 1st ('START') for example.

        Returns
        -------
        {klass}
            The transformed {klass} with the new frequency.

        See Also
        --------
        {other}.asfreq: Convert each Period in a {other_name} to the given frequency.
        Period.asfreq : Convert a :class:`~pandas.Period` object to the given frequency.

        Examples
        --------
        >>> pidx = pd.period_range('2010-01-01', '2015-01-01', freq='A')
        >>> pidx
        PeriodIndex(['2010', '2011', '2012', '2013', '2014', '2015'],
        dtype='period[A-DEC]')

        >>> pidx.asfreq('M')
        PeriodIndex(['2010-12', '2011-12', '2012-12', '2013-12', '2014-12',
        '2015-12'], dtype='period[M]')

        >>> pidx.asfreq('M', how='S')
        PeriodIndex(['2010-01', '2011-01', '2012-01', '2013-01', '2014-01',
        '2015-01'], dtype='period[M]')
        """
        ...
    
    def astype(self, dtype, copy: bool = ...): # -> Self@PeriodArray | PeriodArray | DatetimeArray | NDArray[object_] | ndarray[Any, Unknown] | ExtensionArray | NDArray[Any] | NDArray[int64] | NDArray[Unknown]:
        ...
    
    def searchsorted(self, value: NumpyValueArrayLike | ExtensionArray, side: Literal["left", "right"] = ..., sorter: NumpySorter = ...) -> npt.NDArray[np.intp] | np.intp:
        ...
    
    def fillna(self, value=..., method=..., limit=...) -> PeriodArray:
        ...
    


def raise_on_incompatible(left, right): # -> IncompatibleFrequency:
    """
    Helper function to render a consistent error message when raising
    IncompatibleFrequency.

    Parameters
    ----------
    left : PeriodArray
    right : None, DateOffset, Period, ndarray, or timedelta-like

    Returns
    -------
    IncompatibleFrequency
        Exception to be raised by the caller.
    """
    ...

def period_array(data: Sequence[Period | str | None] | AnyArrayLike, freq: str | Tick | None = ..., copy: bool = ...) -> PeriodArray:
    """
    Construct a new PeriodArray from a sequence of Period scalars.

    Parameters
    ----------
    data : Sequence of Period objects
        A sequence of Period objects. These are required to all have
        the same ``freq.`` Missing values can be indicated by ``None``
        or ``pandas.NaT``.
    freq : str, Tick, or Offset
        The frequency of every element of the array. This can be specified
        to avoid inferring the `freq` from `data`.
    copy : bool, default False
        Whether to ensure a copy of the data is made.

    Returns
    -------
    PeriodArray

    See Also
    --------
    PeriodArray
    pandas.PeriodIndex

    Examples
    --------
    >>> period_array([pd.Period('2017', freq='A'),
    ...               pd.Period('2018', freq='A')])
    <PeriodArray>
    ['2017', '2018']
    Length: 2, dtype: period[A-DEC]

    >>> period_array([pd.Period('2017', freq='A'),
    ...               pd.Period('2018', freq='A'),
    ...               pd.NaT])
    <PeriodArray>
    ['2017', '2018', 'NaT']
    Length: 3, dtype: period[A-DEC]

    Integers that look like years are handled

    >>> period_array([2000, 2001, 2002], freq='D')
    <PeriodArray>
    ['2000-01-01', '2001-01-01', '2002-01-01']
    Length: 3, dtype: period[D]

    Datetime-like strings may also be passed

    >>> period_array(['2000-Q1', '2000-Q2', '2000-Q3', '2000-Q4'], freq='Q')
    <PeriodArray>
    ['2000Q1', '2000Q2', '2000Q3', '2000Q4']
    Length: 4, dtype: period[Q-DEC]
    """
    ...

@overload
def validate_dtype_freq(dtype, freq: BaseOffsetT) -> BaseOffsetT:
    ...

@overload
def validate_dtype_freq(dtype, freq: timedelta | str | None) -> BaseOffset:
    ...

def validate_dtype_freq(dtype, freq: BaseOffsetT | timedelta | str | None) -> BaseOffsetT:
    """
    If both a dtype and a freq are available, ensure they match.  If only
    dtype is available, extract the implied freq.

    Parameters
    ----------
    dtype : dtype
    freq : DateOffset or None

    Returns
    -------
    freq : DateOffset

    Raises
    ------
    ValueError : non-period dtype
    IncompatibleFrequency : mismatch between dtype and freq
    """
    ...

def dt64arr_to_periodarr(data, freq, tz=...) -> tuple[npt.NDArray[np.int64], BaseOffset]:
    """
    Convert an datetime-like array to values Period ordinals.

    Parameters
    ----------
    data : Union[Series[datetime64[ns]], DatetimeIndex, ndarray[datetime64ns]]
    freq : Optional[Union[str, Tick]]
        Must match the `freq` on the `data` if `data` is a DatetimeIndex
        or Series.
    tz : Optional[tzinfo]

    Returns
    -------
    ordinals : ndarray[int64]
    freq : Tick
        The frequency extracted from the Series or DatetimeIndex if that's
        used.

    """
    ...

